# Standard library
from typing import Union

# Third party
from peewee import Field


class Rarity(object):
    """
    Instances of this object represent the rarity of a reward.
    There are four different rarities (listed in ascending magnitude): COMMON, UNCOMMON, RARE, LEGENDARY.
    A rarity object can be created using theses string names or the int indices 1, 2, 3, 4

    CHANGELOG

    Added 09.06.2019
    """
    # CLASS ATTRIBUTES / GLOBAL VARIABLES
    # -----------------------------------

    COMMON = 'common'
    UNCOMMON = 'uncommon'
    RARE = 'rare'
    LEGENDARY = 'legendary'

    RARITIES = [
        COMMON,
        UNCOMMON,
        RARE,
        LEGENDARY
    ]

    DEFAULT_VALUE = COMMON

    # These dicts exist to convert a string rarity representation to its integer format and the other way around.
    # This conversion is for example needed, when an integer is passed to the constructor or, when two rarities are
    # supposed to be compared with GREATER or LOWER
    INTEGER_MAP = {
        1: COMMON,
        2: UNCOMMON,
        3: RARE,
        4: LEGENDARY,
    }
    STRING_MAP = {
        COMMON:     1,
        UNCOMMON:   2,
        RARE:       3,
        LEGENDARY:  4,
    }

    # INSTANCE CONSTRUCTION
    # ---------------------

    def __init__(self, value: Union[str, int]):
        """
        The constructor.

        CHANGELOG

        Added 09.06.2019

        :param value:
        """
        self.value = self.DEFAULT_VALUE

        # Depending on what type of value is being passed to the constructor, we simply pass them along to their
        # individual methods to be processed (modifying the 'value' attribute of the object
        if isinstance(value, str):
            self._init_string(value)
        else:
            self._init_int(int(value))

    def _init_string(self, value: str):
        """
        This method is being called by the constructor, when a string is passed to it, it will check if the string
        has the right format and assign the correct rarity to "value" if it does, otherwise raise a ValueError

        CHANGELOG

        Added 09.06.2019

        :raises: ValueError

        :param value:
        :return:
        """
        # Passing a string as the value should be case insensitive, thus we only use the lower case version of
        # whatever string has been passed
        value_lower = value.lower()
        if value_lower in self.RARITIES:
            self.value = value_lower
        else:
            raise ValueError('The given string "%s" is not a rarity!' % value)

    def _init_int(self, value: int):
        """
        This method is being called by the constructor, when a integer is passed to it, it will check if the int is
        one of the possible rarity id's and assign the according rarity to "self.value", raises ValueError otherwise.

        CHANGELOG

        Added 09.06.2019

        :raises: ValueError

        :param value:
        :return:
        """
        # In the static attribute "INTEGER_MAP" all possible integer values, representing a rarity value are listed as
        # the keys. If a integer is passed that does not equate to a rarity an exception is risen-
        if value in self.INTEGER_MAP.keys():
            self.value = self.INTEGER_MAP[value]
        else:
            raise ValueError('The given integer "%s" is not a rarity!' % value)

    # MAGIC METHODS
    # -------------

    def __str__(self):
        """
        When the str() method is called on the object, it will return the string identifier for the given rarity
        :return:
        """
        return self.value

    def __int__(self):
        """
        When the int() function is called on the object it will return the int id for the given rarity
        :return:
        """
        return self.STRING_MAP[self.value]

    def __eq__(self, other):
        return int(self) == int(Rarity(other))

    def __gt__(self, other):
        return int(self) > int(Rarity(other))

    def __ge__(self, other):
        return int(self) >= int(Rarity(other))

    def __lt__(self, other):
        return int(self) < int(Rarity(other))

    def __le__(self, other):
        return int(self) <= int(Rarity(other))


class RarityField(Field):
    """
    This is a custom field for the database models, which saves a rarity object (most likely to be used as the
    rarity of a reward)

    CHANGELOG

    Added 09.06.2019
    """
    field_type = 'integer'

    def db_value(self, value):
        """
        This method is called when a python model object is to be saved into the database. This method derives the
        integer representation of a given rarity.
        The rarity field can be assigned a string and int identifier of a rarity or a rarity object directly

        CHANGELOG

        Added 09.06.2019

        :param value:
        :return:
        """
        # Coming from the "realm of python" we have to get the integer representation of the rarity object.
        # To enable the usage of setting strings to a rarity field also, we first create a new Rarity object from
        # the passed value and then insert the integer representation of that into the database
        rarity = Rarity(value)
        return int(rarity)

    def python_value(self, value):
        """
        This method is called, when a object is being loaded from the database. It converts the integer representation
        of the rarity into a Rarity object, which is then set as the attribute of the models instance.

        CHANGELOG

        Added 09.06.2019

        :param value:
        :return:
        """
        # We need to convert the value, which is coming from the database into a "Rarity" object to be used in the
        # python program
        return Rarity(value)



# Standard library
from unittest import TestCase

# Third party
from peewee import SqliteDatabase
from peewee import Model
from peewee import CharField

# Local imports
from rewardify.rarity import Rarity

from rewardify.probability import RarityProbability
from rewardify.probability import ProbabilityField


class TestRarityProbability(TestCase):

    SAMPLE_VALUE = 0.25
    SAMPLE_VALUE_LIST = [SAMPLE_VALUE] * 4

    # THE ACTUAL TEST CASES
    # ---------------------

    def test_construction_works(self):
        rarity_probability = RarityProbability(0.25, 0.25, 0.25, 0.25)
        self.assertIsInstance(rarity_probability, RarityProbability)

    def test_indexing_with_rarity_works(self):
        rarity_probability = self.get_sample_object()

        # Now we are testing the various methods to index a RarityProbability object
        self.assertEqual(rarity_probability[1], self.SAMPLE_VALUE)
        self.assertEqual(rarity_probability['common'], self.SAMPLE_VALUE)
        self.assertEqual(rarity_probability[Rarity('common')], self.SAMPLE_VALUE)

    def test_converting_into_list(self):
        rarity_probability = self.get_sample_object()

        probability_list = rarity_probability.list()
        self.assertListEqual(probability_list, self.SAMPLE_VALUE_LIST)

    def test_creating_from_list(self):
        rarity_probability = RarityProbability.from_iterable(self.SAMPLE_VALUE_LIST)

        self.assertEqual(rarity_probability['common'], self.SAMPLE_VALUE)

    def test_creating_csv_string_from_object(self):
        expected_string = '0.25,0.25,0.25,0.25'
        rarity_probability = self.get_sample_object()

        csv_string = rarity_probability.csv(accuracy=2)
        self.assertEqual(csv_string, expected_string)

    def test_creating_object_from_csv_string(self):
        csv_string = '0.5,0.3,0.1,0.1'
        rarity_probability = RarityProbability.from_csv(csv_string)

        self.assertEqual(rarity_probability['common'], 0.5)

    def test_equals(self):
        rarity_probability1 = RarityProbability(0.25, 0.25, 0.25, 0.25)
        rarity_probability2 = RarityProbability(0.25, 0.25, 0.25, 0.25)
        rarity_probability3 = RarityProbability(0.7, 0.1, 0.1, 0.1)

        # objects 1 and 2 must be equal. Object 3 is different
        self.assertEqual(rarity_probability1, rarity_probability2)
        self.assertNotEqual(rarity_probability1, rarity_probability3)

    def test_choice(self):
        rarity_probability = RarityProbability(0.25, 0.25, 0.25, 0.25)
        result = rarity_probability.choice()
        self.assertTrue(result in Rarity.RARITIES)

    # HELPER METHODS
    # --------------

    def get_sample_object(self):
        return RarityProbability(self.SAMPLE_VALUE, self.SAMPLE_VALUE, self.SAMPLE_VALUE, self.SAMPLE_VALUE)


# TESTING THE PROBABILITY FIELD
# -----------------------------


class ProbabilityModel(Model):

    name = CharField()
    slot = ProbabilityField()


class TestProbabilityField(TestCase):
    TEST_DATABASE = SqliteDatabase(':memory:')
    MODELS = [ProbabilityModel]

    SAMPLE_MODEL_NAME = 'sample'
    SAMPLE_RARITY_PROBABILITY = RarityProbability(0.25, 0.25, 0.25, 0.25)

    # SETTING UP THE TEST DATABASE
    # ----------------------------

    def setUp(self):
        self.TEST_DATABASE.bind(self.MODELS, bind_refs=False, bind_backrefs=False)

        self.TEST_DATABASE.connect()
        self.TEST_DATABASE.create_tables(self.MODELS)

    def tearDown(self):
        self.TEST_DATABASE.drop_tables(self.MODELS)
        self.TEST_DATABASE.close()

    # ACTUAL TEST CASES
    # -----------------

    def test_model_creation_works(self):
        # Saving the model into the database
        model = ProbabilityModel(
            name=self.SAMPLE_MODEL_NAME,
            slot=RarityProbability(0.25, 0.25, 0.25, 0.25)
        )
        model.save()

        # Loading the model again from the database
        model = ProbabilityModel.get(ProbabilityModel.name == self.SAMPLE_MODEL_NAME)
        self.assertIsInstance(model.slot, RarityProbability)

    def test_model_creation_works_with_list(self):
        model = ProbabilityModel(
            name=self.SAMPLE_MODEL_NAME,
            slot=[0.5, 0.2, 0.2, 0.1]
        )
        model.save()

        model = self.get_sample_model()
        self.assertEqual(model.slot['common'], 0.5)
        self.assertEqual(model.slot['legendary'], 0.1)

    # HELPER METHODS
    # --------------

    def get_sample_model(self):
        return ProbabilityModel.get(ProbabilityModel.name == self.SAMPLE_MODEL_NAME)



# Third party imports
from argon2 import PasswordHasher
from argon2.exceptions import VerifyMismatchError

from peewee import Field


class PasswordHash(str):
    """
    This class inherits from a normal "string" object.
    It represents a password hash. Only hashs are being written to the database, so not to be able to possibly leak
    plain text passwords.
    The password string gets passed to the constructor, but is then immediately hashed and discarded, so only the hash
    string is saved.

    Objects of the PasswordHash type provide direct method access to check if a given password string
    correlates to the hash.

    CHANGELOG

    Added 08.06.2019
    """

    # CLASS VARIABLES
    # ---------------

    # The password hasher object, which we need to hash the password strings for us
    PASSWORD_HASHER = PasswordHasher()

    # This is a substring, which every PasswordHash contains to identify it as a string.
    # we will use a "string" contains to check if a given string is already a hash or a plain text password
    HASH_IDENTIFIER = "$argon2id$"

    # OVERWRITING STRING CONSTRUCTOR
    # ------------------------------

    def __new__(cls, password_string: str):
        """
        The __new__ method is called when a new object is being instantiated from the class. In this case we use it to
        modify the construction logic of the str parent class to implement custom logic:
        When the string already is a hash, it will be set as the value of the string. If it is detected, that the given
        string is still a plain text password though, it will first be hashed.

        CHANGELOG

        Added 08.06.2019

        Changed 09.06.2019
        Added the distinction between the string already being a hash or a plain text password.

        :param password_string:
        :return:
        """
        # IMPORTANT NOTICE:
        # In order to overwrite the value within a class, that inherits from the built in "str" class (as would be the
        # case when some preprocessing or modification of the string has to be made) the __new__ method has to be used!
        # The __init__ method will not work!

        # If the string, that is being passed to the constructor already is the string representation of a hash we just
        # use that value, if not we need to hash it first!
        if cls.is_hash(password_string):
            password_hash = password_string
        else:
            password_hash = cls.PASSWORD_HASHER.hash(password_string)
        return str.__new__(cls, password_hash)

    def __init__(self, password_string: str):
        super(PasswordHash, self).__init__()

    # UTILITY METHODS
    # ---------------

    def check(self, password_string: str) -> bool:
        """
        Given a password string, that was entered by a user through the front end, this method will check if the
        given password string correlates with this very hash and return true if it does so.

        CHANGELOG

        Added 08.06.2019

        :param password_string:
        :return:
        """
        try:
            self.PASSWORD_HASHER.verify(self, password_string)
            return True
        except VerifyMismatchError:
            return False

    def verify(self, password_string: str):
        """
        Given a password string, that was entered by a user through the front end, this method will check if the
        given password string correlates with this very hash and raises an exception if the given string does not
        match with the hash

        CHANGELOG

        Added 08.06.2019

        :param password_string:
        :return: void
        """
        self.PASSWORD_HASHER.verify(self, password_string)

    @classmethod
    def is_hash(cls, string):
        """
        Given a string, this method returns true, if the given string is already the hash of a password and false if
        it still is a plain text password.

        CHANGELOG

        Added 08.06.2019

        :param string:
        :return: bool
        """
        if cls.HASH_IDENTIFIER in string:
            return True
        else:
            return False


class PasswordField(Field):
    """
    This is a custom field for database models, which stores a password in the form of a hash. The plain text password
    will not be stored!

    CHANGELOG

    Added 09.06.2019
    """

    field_type = 'text'

    # FIELD ACCESS METHODS
    # --------------------

    def db_value(self, value):
        """
        This method is called when the python value has to be converted into a actual database value, for that we will
        use the hashed string representation of the password string

        CHANGELOG

        Added 09.06.2019

        :param value:
        :return:
        """
        # Given a value from the "realm of python code execution" this method has to return a value, that will
        # ultimately be passed to the database to be saved into a (string) field.
        password_hash = PasswordHash(value)
        return str(password_hash)

    def python_value(self, value):
        """
        This method is called when the database value is supposed to be converted into a python object, which will then
        serve as the attribute to the model object.
        In this case we will take the string hash and convert it into a PasswordHash object, as that features
        special methods to check for password validity

        CHANGELOG

        Added 09.06.2019

        :param value:
        :return:
        """
        # This method is given the raw value, which is retrieved from the database and has the job of converting
        # it into the correct python object.
        # This would be a "PasswordHash" instance
        password_hash = PasswordHash(value)
        return password_hash



# Standard library
import datetime

from typing import List, Dict, Any


# ###########################################
# THE INTERFACE FOR ALL BACKENDS TO IMPLEMENT
# ###########################################


class AbstractBackend:
    """
    This is the abstract base class, from which any concrete backend implementation has to inherit.
    It defines the simple "get_update" method, which has to be implemented to return the actions for which the
    users have earned gold.

    Each successive call to get_update should only return those actions, that are NEW since the last update, which
    means a backend has to keep track of which actions have already been processed and which were not.
    """

    def get_update(self) -> Dict[str, List]:
        """
        This is the main and only interface a concrete implementation of a gold providing backend has to satisfy. A
        call to this method should return a dictionary, where the key is the username of the user to update. The value
        to that username is supposed to be a list of dicts, where each dict describes an action, that has given the
        user the gold. The dicts, which describe these actions should contain the following keys:
        - name: Name of the action, that has granted the gold
        - description: Short description of the action
        - gold: The amount of gold, that was earned
        - date: A datetime object for the date, at which the action was completed

        :return:
        """
        raise NotImplementedError()


# #######################
# USEFUL MIXIN BEHAVIOURS
# #######################

class CheckConfigMixin:
    """
    This mixin provides the "check_config" function, which will go through all the variable names, the backend requires
    for correct functionality and checks of the EnvironmentConfig actually contains those.

    This mixin expects two things:
    - The class must implement the "get_config" method, which returns the EnvironmentConfig instance
    - The class must implement the "expected_config" method, which returns a list with all the variable names, that are
    required for the backend

    CHANGELOG

    Added 13.06.2019
    """

    def check_config(self):
        """
        This method will check if all the variables expected from the config file are actually defined in the config.
        self.EXPECTED_CONIFG contains the list with all the variable names.
        In case not all are defined a ModuleNotFoundError will be risen.

        CHANGELOG

        Added 10.06.2019

        Changed 13.06.2019
        This was originally only a method of the mock backend, but I found it so useful for other backends as well,
        that I made it it's own mixin

        .:raises: ModuleNotFoundError

        :return:
        """
        config = self.get_config()
        for variable in self.expected_config():
            if not hasattr(config, variable):
                raise ModuleNotFoundError("The config file is missing the %s variable!" % variable)

    def get_config(self):
        """
        Returns a reference to the EnvironmentConfig instance

        CHANGELOG

        Added 13.06.2019

        :return:
        """
        raise NotImplementedError()

    def expected_config(self):
        """
        Returns a list with the string names of the expected variables for the backend

        CHANGELOG

        Added 13.06.2019

        :return:
        """
        raise NotImplementedError()



import os


# #########
# CONSTANTS
# #########

PATH = os.path.dirname(os.path.abspath(__file__))

# ########################
# BUSINESS LOGIC FUNCTIONS
# ########################


def rarity_filter(rarity: str):
    """
    This is a decorator, which will return a function, that can be used as the callable for the filter() function when
    filtering the available rewards defined in the config. The function will filter True for every reward which was
    defined with the given rarity

    CHANGELOG

    Added 12.06.2019

    :param rarity:
    :return:
    """
    def _func(reward: dict):
        return reward['rarity'] == rarity

    return _func

# ####################
# PROGRAMMING PATTERNS
# ####################


class Singleton:
    """
    A non-thread-safe helper class to ease implementing singletons.
    This should be used as a decorator -- not a metaclass -- to the
    class that should be a singleton.

    The decorated class can define one `__init__` function that
    takes only the `self` argument. Also, the decorated class cannot be
    inherited from. Other than that, there are no restrictions that apply
    to the decorated class.

    To get the singleton instance, use the `instance` method. Trying
    to use `__call__` will result in a `TypeError` being raised.

    """

    def __init__(self, decorated):
        self._decorated = decorated

    def instance(self):
        """
        Returns the singleton instance. Upon its first call, it creates a
        new instance of the decorated class and calls its `__init__` method.
        On all subsequent calls, the already created instance is returned.

        """
        try:
            return self._instance
        except AttributeError:
            self._instance = self._decorated()
            return self._instance

    def __call__(self):
        raise TypeError('Singletons must be accessed through `instance()`.')

    def __instancecheck__(self, inst):
        return isinstance(inst, self._decorated)



# Standard library
import platform
import os
import importlib.util
import ast
import logging

from typing import NoReturn, Dict

# Third party
from peewee import SqliteDatabase, MySQLDatabase, PostgresqlDatabase

from jinja2 import FileSystemLoader, Environment

# Local imports
from rewardify.models import User, Reward, Pack, DATABASE_PROXY

from rewardify._util import Singleton, PATH


FOLDER_PATH = os.path.dirname(os.path.abspath(__file__))


# #################################
# DATABASE CONFIGURATION MANAGEMENT
# #################################


class DatabaseConfig:
    """
    Instances of this class manage the configuration of tha database for the project. This includes the engine, host,
    port, username, password and database name.

    The typical work flow using this class is the following:
    A new object is created from a config dict. Then the database is init-ed (meaning the object is created)
    Then the database proxy (placeholder) for the models is connected to this database object at run time and then the
    database is actually connected to its persistent counterpiece (file for sqlite, server for mysql):

    EXAMPLE:
    database_config = DatabaseConfig.from_dict(database_dict)
    database_config.init()
    database_proxy.initialize(database_config.get())
    database_config.connect()
    database_config.create_tables() # OPTIONAL

    CHANGELOG

    Added 09.06.2019
    """
    # This dict will be used as the default base dict, when a new DatabaseConfig object is to be created from a dict,
    # which may not contain all the keys. For the keys that are not present, the default ones will stay, but will
    # otherwise be overwritten.
    DEFAULT_DICT = {
        'engine':       'sqlite',
        'host':         ':memory:',
        'database':     '',
        'user':         '',
        'password':     '',
        'port':         0
    }

    # A list of all valid engine strings
    SUPPORTED_ENGINES = [
        'sqlite',
        'mysql',
        'postgres'
    ]

    # This is a list, that contains all the model classes of this project. This list is needed to be passed
    # to the database to create all the tables
    MODELS = [
        User,
        Reward,
        Pack
    ]

    # INSTANCE CONSTRUCTION
    # ---------------------

    def __init__(self,
                 engine: str,
                 host: str,
                 database: str,
                 user: str,
                 password: str,
                 port: int) -> NoReturn:
        # This will raise an error, if the engine string is not a valid one
        self.check_engine(engine)

        self.engine = engine
        self.host = host
        self.database = database
        self.user = user
        self.password = password
        self.port = port

        # Later on this variable will contain the actual database instance
        self.instance = None

    def check_engine(self, engine: str):
        """
        This will raise a value error if the engine string is not one of the supported engines.

        CHANGELOG

        Added 09.06.2019

        :param engine:
        :return:
        """
        if engine not in self.SUPPORTED_ENGINES:
            raise ValueError('The engine "{}" is not supported'.format(engine))

    # CREATING DATABASE INSTANCE
    # --------------------------

    def init(self):
        """
        This method will create a new database object and put it into the "self.instance" variable. The type of
        database object created depends on the value of the engine string

        CHANGELOG

        Added 09.06.2019

        :return:
        """
        # Basically I am using this dict as a "switch" statement to reduce "if"-clutter
        engine_create_methods = {
            'sqlite':       self.init_sqlite,
            'mysql':        self.init_mysql,
            'postgres':     self.init_postgres
        }
        engine_create_methods[self.engine]()

    def init_sqlite(self):
        """
        Creates a new sqlite database object and puts it into "self.instance"
        NOTE: For the creation of a sqlite database only the path to the file is relevant. The "self.host" variable
        will be used as the sqlite string

        CHANGELOG

        Added 09.06.2019

        :return:
        """
        self.instance = SqliteDatabase(self.host)

    def init_mysql(self):
        """
        Creates a new mysql database object and puts it into the "self.instance".

        CHANGELOG

        Added 09.06.2019

        :return:
        """
        self.instance = MySQLDatabase(
            self.database,
            host=self.host,
            user=self.user,
            password=self.password,
            port=self.port
        )

    def init_postgres(self):
        """
        Creates a new postgrsql database object and puts it into the "self.instance"

        CHANGELOG

        Added 09.06.2019

        :return:
        """
        self.instance = PostgresqlDatabase(
            self.database,
            host=self.host,
            user=self.user,
            password=self.password,
            port=self.port
        )

    # DATABASE OPERATIONS
    # -------------------

    def connect(self):
        """
        When a database object has already been created, this method actually connects the object in "self.instance"
        to the actual persistent database.

        CHANGELOG

        Added 09.06.2019

        :return:
        """
        self.instance.connect()

    def create_tables(self):
        """
        If the database object is already connected and created, this method will create the tables for all the models
        of the rewardify project into this database.

        CHANGELOG

        Added 09.06.2019

        :return:
        """
        self.instance.create_tables(self.MODELS)

    def drop_tables(self):
        """
        If the database object is already connected and created, this method will drop the tables for all the models
        of the rewardify project into this database.

        CHANGELOG

        Added 09.06.2019

        :return:
        """
        self.instance.drop_tables(self.MODELS)

    def get(self):
        """
        returns the database object

        CHANGELOG

        Added 09.06.2019

        :return:
        """
        return self.instance

    # UTILITY METHODS
    # ---------------

    def update(self, database_dict: Dict) -> NoReturn:
        """
        Given a database dict, all the internal values will be updated according to this dict

        CHANGELOG

        Added 09.06.2019

        :param database_dict:
        :return:
        """
        for key, value in database_dict.items():
            setattr(self, key, value)

    # CLASS METHODS
    # -------------

    @classmethod
    def from_dict(cls, database_dict):
        """
        Given a dict, which contains the database config, this method will create and return a new DatabaseConfig
        object from it.
        The dict can contain any subset of the following keys, every key not used it substituted with the default item:
        - engine
        - host
        - database
        - port
        - user
        - password

        CHANGELOG

        Added 09.06.2019

        :param database_dict:
        :return:
        """
        # We will use the default dict as a base line (in case not all keys are provided by the given dict as for the
        # case of sqlite. Then we replace all possible values with the actual ones
        argument_dict = cls.DEFAULT_DICT
        argument_dict.update(database_dict)

        # Finally we can pass the arguments to the constructor by unpacking the dict
        return cls(**argument_dict)


# ######################
# CONFIG FILE MANAGEMENT
# ######################


@Singleton
class EnvironmentConfig:

    # CLASS VARIABLES
    # ---------------

    PLATFORM_FOLDER_PATHS = {
        'Linux': '~/.rewardify',
        'Darwin': 'NONE',
        'Windows': 'NONE'
    }

    CONFIG_FILE_NAME = 'config.py'

    TEMPLATE_FOLDER_PATH = os.path.join(FOLDER_PATH, 'templates')
    TEMPLATE_LOADER = FileSystemLoader(searchpath=TEMPLATE_FOLDER_PATH)
    TEMPLATE_ENVIRONMENT = Environment(loader=TEMPLATE_LOADER)

    # THE DEFAULT VALUES
    # ------------------

    DEFAULT_DATABASE_DICT = {}

    DEFAULT_BACKEND_CLASS = None

    def __init__(self):
        # First of all, we need to know in which folder the config is even located in. This is dependant on the
        # platform, but inconfigureable for each platform.
        # We use the mapping of the platform to the folder path
        self.platform = platform.system()
        print(self.PLATFORM_FOLDER_PATHS)
        self.folder_path = self.PLATFORM_FOLDER_PATHS[self.platform]

        # Creating a new database configuration object from the default
        self.database_config = DatabaseConfig.from_dict(self.DEFAULT_DATABASE_DICT)

        self.backend_class = self.DEFAULT_BACKEND_CLASS

    def set_folder_path(self, folder_path):
        self.folder_path = folder_path

    def load(self):
        self.load_config()
        self.load_database()

    def load_config(self):
        config_file_path = os.path.join(self.folder_path, self.CONFIG_FILE_NAME)

        # Here we are dynamically importing the config python module using its file path
        spec = importlib.util.spec_from_file_location('config', config_file_path)
        config = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(config)

        source = self.read_file(config_file_path)

        # Now we parse the module to obtain all the variable names
        root = ast.parse(source, filename=config_file_path)
        names = sorted({node.id for node in ast.walk(root) if isinstance(node, ast.Name)})

        for name in names:
            setattr(self, name, getattr(config, name))

    def load_database(self):
        self.database_config = DatabaseConfig.from_dict(self.DATABASE)

    def read_file(self, file_path: str):
        with open(file_path, mode='r+') as file:
            content = file.read()
        return content

    # CHECKING METHODS
    # ----------------

    def config_folder_exists(self) -> bool:
        """
        Returns True if the config folder path exists, False otherwise

        CHANGELOG

        Added 09.06.2019

        :return:
        """
        return os.path.exists(self.folder_path) and os.path.isdir(self.folder_path)


# ###############################
# INSTALLATION OF THE ENVIRONMENT
# ###############################


class EnvironmentInstaller:
    """
    This is a class, that is supposed to help with the creation of the config environment during the installation
    process.
    This includes the creation of the folder, which will contain all the files relevant to rewardify, the creation of
    a sample config.py file and the creation of a rewardify.db sqlite file, which is already set up to contain the
    needed tables

    CHANGELOG

    Added 14.06.2019
    """

    PLATFORM_CONFIG_FOLDER_NAME = {
        'Linux':            '.rewardify',
        'Darwin':           'rewardify',
        'Windows':          'Rewardify'
    }

    DATABASE_FILE_NAME = "rewardify.db"

    CONFIG_FILE_NAME = "config.py"
    CONFIG_FILE_TEMPLATE = "config.jinja2"

    # CONFIGURATION FOR JINJA TEMPLATING
    # ----------------------------------

    TEMPLATE_FOLDER_PATH = os.path.join(PATH, 'templates')
    TEMPLATE_LOADER = FileSystemLoader(searchpath=TEMPLATE_FOLDER_PATH)
    TEMPLATE_ENVIRONMENT = Environment(loader=TEMPLATE_LOADER)

    # INSTANCE CONSTRUCTION
    # ---------------------

    def __init__(self, install_path: str):
        """
        The constructor.

        CHANGELOG

        Added 14.06.2019

        :param install_path:
        """
        self.logger = logging.getLogger('Installation')
        self.platform = platform.system()

        self.path = install_path

    # INSTALLATION PROCESS
    # --------------------

    def install(self):
        """
        This method will perform the install of the envrionment: Create the folder, the "config.py" and the
        "rewardify.db"

        CHANGELOG

        Added 14.06.2019

        :return:
        """
        self.create_config_folder()
        self.create_config_file()
        self.create_database_file()

    def create_database_file(self):
        """
        Creates the database file "rewardify.db" and creates the necessary tables into the sqlite database

        CHANGELOG

        Added 14.06.2019

        :return:
        """
        database_file_path = self.get_database_file_path()
        self.create_empty_file(database_file_path)

        database = SqliteDatabase(database_file_path)
        DATABASE_PROXY.initialize(database)
        database.connect()
        database.create_tables([User, Pack, Reward])
        database.close()

    def create_config_file(self):
        config_file_path = self.get_config_file_path()
        self.create_file_from_template(
            config_file_path,
            self.CONFIG_FILE_TEMPLATE,
            {}
        )

    def create_config_folder(self):
        config_folder_path = self.get_config_folder_path()
        os.makedirs(config_folder_path)
        self.logger.info('Created config folder at PATH "{}"'.format(config_folder_path))

    # HELPER METHODS
    # --------------

    def get_config_folder_path(self):
        config_folder_name = self.PLATFORM_CONFIG_FOLDER_NAME[self.platform]
        return os.path.join(self.path, config_folder_name)

    def get_database_file_path(self):
        config_folder_path = self.get_config_folder_path()
        return os.path.join(config_folder_path, self.DATABASE_FILE_NAME)

    def get_config_file_path(self):
        config_folder_path = self.get_config_folder_path()
        return os.path.join(config_folder_path, self.CONFIG_FILE_NAME)

    @classmethod
    def create_file_from_template(cls, file_path: str, template_name: str, context: Dict):
        """
        Given the string path of a file to be written to, the name of the template (within the defined template folder)
        and a dict defining the context of for the jinja template, this method will render the template with the
        context and save the result into the given file path.

        CHANGELOG

        Added 10.06.2019

        :param file_path:
        :param template_name:
        :param context:
        :return:
        """
        # First we render the template content into a plain string and then we save the string into the
        # file (overwriting all the content)
        template = cls.TEMPLATE_ENVIRONMENT.get_template(template_name)
        content = template.render(**context)

        with open(file_path, mode='w+') as file:
            file.write(content)

    @classmethod
    def create_empty_file(cls, file_path: str):
        """
        Given a file path, this method will create that file empty. If the file already exists, it will override the
        content with an empty string

        CHANGELOG

        Added 14.06.2019

        :param file_path:
        :return:
        """
        with open(file_path, mode="w+") as file:
            file.write("")



# standard library
import datetime
import os
import re
import json

from typing import Dict, List
from collections import defaultdict

# third party
import pandas as pd

# local

from rewardify.env import EnvironmentConfig

from rewardify.backends.base import AbstractBackend, CheckConfigMixin


class ForestCSVFile:
    """
    This object represents one of the CSV files from the forest app.

    CHANGELOG

    Added 13.06.2019
    """

    COLUMN_NAMES = [
        'Start Time',
        'End Time',
        'Tag',
        'Note',
        'Tree Type',
        'Is Success'
    ]

    # The pandas function, which parses the CSV takes a dict, which assigns data types to the columns. Here we make
    # sure that the "Is Success" column is parsed as boolean values
    PANDAS_DATA_TYPES = {
        'Start Time':       str,
        'End Time':         str,
        'Tag':              str,
        'Note':             str,
        'Tree Type':        str,
        'Is Success':       bool
    }

    # The list of the column names to be parsed as datetime objects has to be passed separately to the
    # function, which parses the CSV file.
    PANDAS_DATETIME_COLUMNS = [
        'Start Time',
        'End Time'
    ]

    def __init__(self, path: str):
        """
        Constructor.

        CHANGELOG

        Added 13.06.2019

        :param path:
        """
        self.path = path
        self.content = self.load()

        # Tha pandas utility for csv parsing is superior to the std "csv" module, as it has auto detection of column
        # names and datatype parsing.
        # A pandas "DataFrame" object makes it significantly easier to handle the data as well, by providing a query
        # method for example.
        self.data_frame = pd.read_csv(
            path,
            dtype=self.PANDAS_DATA_TYPES,
            parse_dates=self.PANDAS_DATETIME_COLUMNS
        )

    def load(self):
        """
        Opens the file specified by "self.path" and returns the string content

        CHANGELOG

        Added 13.06.2019

        :return:
        """
        with open(self.path, "r+") as file:
            return file.read()

    def get_dicts_after(self, date_time: datetime.datetime) -> List[Dict]:
        """
        Returns a list of dicts, where the dicts represent one activity within the csv file. Although only those
        activities are represented, whose starting time is bigger than the given date time (so the ones that have
        occurred after).
        Each dict contains the two keys:
        - duration: The integer duration of the activity
        - tag: the tag of the tree

        :param date_time:
        :return:
        """
        data_frame = self.data_frame.loc[self.data_frame['Start Time'] > date_time]
        return self.get_dicts_from_dataframe(data_frame)

    def get_dicts(self) -> List[Dict]:
        """
        Returns a list of dicts, with one dict per activity / tree within the csv file. Each dict contains the two keys:
        - duration: The integer duration in minutes
        - tag: The tag of the tree

        CHANGELOG

        Added 13.06.2019

        :return:
        """
        return self.get_dicts_from_dataframe(self.data_frame)

    # HELPER METHODS
    # --------------

    @classmethod
    def get_dicts_from_dataframe(cls, data_frame: pd.DataFrame) -> List[Dict]:
        """
        Given a data frame, that has been derived from one of forests CSV files, this method extracts a list of dicts
        from it, where the dicts offer an alternative light weight representation of the activity. Each dict represents
        one tree / activity and only contains two keys "duration" for the duration in minutes and the string "tag".

        CHANGELOG

        Added 13.06.2019

        :param data_frame:
        :return:
        """
        tree_dicts = []
        for index, row in data_frame.iterrows():
            duration = cls.calculate_duration(row['Start Time'], row['End Time'])
            tree_dict = {
                'tag': row['Tag'],
                'duration': duration
            }
            tree_dicts.append(tree_dict)

        return tree_dicts

    @classmethod
    def calculate_duration(cls, start_time, end_time) -> int:
        """
        Given a start time and a end time (as pandas Datetime objects) this method will return the duration between the
        two points in time as an integer value for the minutes.

        CHANGELOG

        Added 13.06.2019

        :param start_time:
        :param end_time:
        :return:
        """
        # Using the subtraction on two Datetime objects creates a "Timedelta" object, which does not contain a direct
        # porperty "minutes", only "seconds". So the minutes have to be calculated from the seconds first
        time_delta: pd.Timedelta = end_time - start_time
        minutes = int(time_delta.seconds / 60)
        return minutes


class ForestBackend(CheckConfigMixin, AbstractBackend):

    EXPECTED_CONFIG = [
        'FOREST_BACKEND_PATH',
        'FOREST_BACKEND_SETTINGS'
    ]

    # This is the name of the file, which will contain the json information about, where the last update process of
    # the backend has left of, aka the info about which trees are new and for which gold has already been granted
    STATE_FILE_NAME = 'last_trees.json'

    # INSTANCE CONSTRUCTION
    # ---------------------

    def __init__(self):
        """
        The constructor.

        CHANGELOG

        Added 13.06.2019
        """
        # First we need an instance of the config, because the config contains all the configuration for the backend
        # as well.
        self.config: EnvironmentConfig = EnvironmentConfig.instance()
        # This method will check if all the variables from "expected_config()" are actually in the config environment
        # and raises a ModuleNotFoundError if not. After it we can assume the variables exist
        self.check_config()

        self.folder_path = self.config.FOREST_BACKEND_PATH
        # This will be a dict, whose keys are the user names from rewardify and the values are dicts, which define
        # their forest user name and their gold preferences for different tree tags.
        self.user_settings = self.config.FOREST_BACKEND_SETTINGS

        # This method will iterate the base folder and find all the relevant csv files.
        self.csv_files: List[ForestCSVFile] = []
        self.load_csv_files()

        # This method will create a dict, whose keys are the user names for the forest app and the values are the
        # user names for rewardify
        self.forest_user_map = self.get_forest_user_mapping()

        self.state = self.load_state()

    def load_csv_files(self):
        """
        This method will walk through the folder specified in "self.folder_path" and create a new ForestCSVFile object
        from each ".csv" file it finds, which it then appends to the "self.csv_files" list.

        CHANGELOG

        Added 13.06.2019

        :return:
        """
        self.csv_files = []
        for root, dirs, files in os.walk(self.folder_path):
            for file in files:
                if '.csv' in file:
                    path = os.path.join(root, file)
                    # The ForestCSVFile is a wrapper class, which simplifies the csv parsing and provides easy access
                    # methods for the information that are most relevant to the backend
                    csv_file = ForestCSVFile(path)
                    self.csv_files.append(csv_file)
            # Without a break here "os.walk" would start to iterate through sub folders as well
            break

    def load_state(self):
        """
        This function loads the state file from the file system and returns the state dict. The state dict is a dict,
        which contains the string names of the rewardify users as the keys and the pandas Timestamp objects for the
        last tree, that has been accounted for during the previous run of the backend as values.

        CHANGELOG

        Added 13.06.2019

        :return:
        """
        # If the file does not already exists, we are going to create it first
        file_path = self.get_state_file_path()
        if not os.path.exists(file_path):
            self.save_state(self._default_state())

        with open(file_path, mode='r') as file:
            state = json.load(file, object_hook=self._date_hook)

        return state

    # HELPER METHODS
    # --------------

    def calculate_gold(self, username: str, tag: str, duration: int) -> int:
        """
        Given the username, whose settings to be used, the tag string and the duration of the activity in minutes, this
        method will calculate the gold, which has been earned for the activity.

        CHANGELOG

        Added 13.06.2019

        :param username:
        :param tag:
        :param duration:
        :return:
        """
        # The user has specified "period" in the settings dict, this is the timespan in minutes, for which he wants to
        # get all the gold values, that he has specified.
        # So by getting the relation of the actual duration to this period we get the multiplier with which to multiply
        # the base value of the gold
        period = self.user_settings[username]['period']
        gold_multilpier = duration / period

        # The user only specifies the gold value for some possible tags, all other tags will give the default amount
        # by creating a default dict we can avoid the user of an if statement, because if the tag is not a key, the
        # default value will be returned
        tag_gold_map = defaultdict(lambda: self.user_settings[username]['default'])
        tag_gold_map.update(self.user_settings[username])

        base_gold = tag_gold_map[tag]
        gold = int(base_gold * gold_multilpier)
        return gold

    def get_forest_user_mapping(self):
        """
        Returns a dict, which maps the user names of rewardify to the user names for the forest app. The keys of the
        dict will be the strin user names for forest and the values the string user names for rewardify

        CHANGELOG

        Added 13.06.2019

        :return:
        """
        forest_user_map = {}
        for username, settings in self.user_settings.items():
            forest_name = settings['name']
            forest_user_map[forest_name] = username
        return forest_user_map

    def get_new_user_activities(self) -> Dict:
        """
        This method returns a dict, whose keys are the user names of the rewardify users and the values are lists with
        activity dicts associated with those users. An activity dict has two keys: "tag" for the kind of tree that was
        planted and "duration" for the time in minutes the activity lasted.
        The activities are new in the sense, that the lists only contain dicts for those trees that have been planted
        AFTER the last tree, that has been processed during the previous run of the backend

        CHANGELOG

        Added 13.06.2019

        :return:
        """
        user_activities = {}
        user_csv_files = self.get_user_csv_files()
        for user_name, csv_files in user_csv_files.items():
            activities = []
            for csv_file in csv_files:  # type: ForestCSVFile
                # From the state dict we get the pandas Timestamp object, which marks the starting time of the last
                # activity, that has been accounted for.
                # Thus we can get the new actovities as all the activities AFTER that last one
                last_start_time = self.state[user_name]
                activities += csv_file.get_dicts_after(last_start_time)

            user_activities[user_name] = activities

        return user_activities

    def get_user_csv_files(self):
        """
        Returns a dict, whose keys are the user names of the rewardify users and the values are the ForestCSVFile
        objects associated with them.

        CHANGELOG

        Added 13.06.2019

        :return:
        """
        user_csv_files = defaultdict(list)
        for csv_file in self.csv_files:
            forest_name = self.forest_name_from_path(csv_file.path)
            user_name = self.forest_user_map[forest_name]
            user_csv_files[user_name].append(csv_file)
        return user_csv_files

    def get_state_file_path(self) ->str:
        """
        Returns the string file path for the state file. This is the json file, which contains the date of the last
        tree, that has been accounted for during the last run of the backend.

        CHANGELOG

        Added 13.06.2019

        :return:
        """
        file_path = os.path.join(self.folder_path, self.STATE_FILE_NAME)
        return file_path

    def save_state(self, state: Dict):
        """
        Given the state dict, this function will save it as the state json file.

        CHANGELOG

        Added 13.06.2019

        :param state:
        :return:
        """
        file_path = self.get_state_file_path()
        with open(file_path, mode="w") as file:
            # The self._str_hook method calls the str() function on all the values of the dict and then returns it
            # again. This is important, because the pandas Timestamp objects are not in themselves json serializable
            json.dump(self._str_hook(state), file)

    @classmethod
    def forest_name_from_path(cls, path):
        """
        Given the path string to a forest csv file, this method will extract the forest user name from that string

        CHANGELOG

        Added 13.06.2019

        :param path:
        :return:
        """
        file_name = os.path.basename(path)
        # This regex pattern extracts the string, which is between two brackets
        result = re.search(".*\((.+?)\).*", file_name)
        forest_name = result.group(1)
        return forest_name

    @classmethod
    def _date_hook(cls, json_dict: Dict) -> Dict:
        """
        This is a function, which takes a dict and attempts to convert all the values into pandas Timestamp objects
        and then returns the dict again.
        A function of this type (takes a dict, gives a dict) can be passed to the json.load() function to make a type
        conversion.

        CHANGELOG

        Added 13.06.2019

        :param json_dict:
        :return:
        """
        for key, value in json_dict.items():
            json_dict[key] = pd.to_datetime(value)
        return json_dict

    @classmethod
    def _str_hook(cls, json_dict: Dict) -> Dict:
        """
        This is a function, which takes a dict and attempts to convert all the values into strings and then returns
        teh dict again.
        A function of this type (takes a dict, gives a dict) can be passed to the json.load() function to make a type
        conversion.

        CHANGELOG

        Added 13.06.2019

        :param json_dict:
        :return:
        """
        for key, value in json_dict.items():
            json_dict[key] = str(value)
        return json_dict

    def _default_state(self):
        """
        Returns the default state dict, which dates the last successful tree at 1701.

        CHANGELOG

        Added 13.06.2019

        :return:
        """
        last_trees = {}
        for user_name in self.user_settings.keys():
            last_trees[user_name] = pd.to_datetime("Sat Jan 15 08:07:54 GMT+01:00 1701")
        return last_trees

    # IMPLEMENTING INTERFACE
    # ----------------------

    def get_update(self):
        """
        This is the main function for the backend, it will return a dict, whose keys are the user names and whose
        values are lists of dicts, which each describe the gold, that has been earned for a certain activity.

        CHANGELOG

        Added 13.06.2019

        :return:
        """
        update_dict = defaultdict(list)
        user_activities = self.get_new_user_activities()
        for user_name, activities in user_activities.items():
            for activity in activities:
                gold = self.calculate_gold(user_name, activity['tag'], activity['duration'])
                action_dict = {
                    'name':         'Tree',
                    'description':  activity['tag'],
                    'gold':         gold,
                    'date':         datetime.datetime.now()
                }
                update_dict[user_name].append(action_dict)

        return update_dict

    # IMPLEMENTING MIXIN REQUIREMENTS
    # -------------------------------

    def get_config(self):
        """
        Returns a reference to the EnvironmentConfig instance

        CHANGELOG

        Added 13.06.2019

        :return:
        """
        return self.config

    def expected_config(self):
        """
        Returns a list with the string names of the expected config variables.

        CHANGELOG

        Added 13.06.2019

        :return:
        """
        return self.EXPECTED_CONFIG



# standard library
import re

from typing import Dict, List

from collections import defaultdict

# Third party imports
import numpy as np

from peewee import DatabaseProxy
from peewee import Model
from peewee import ForeignKeyField, CharField, IntegerField, DateTimeField, TextField

# Local imports
from rewardify._util import rarity_filter

from rewardify.password import PasswordField

from rewardify.rarity import Rarity
from rewardify.rarity import RarityField

from rewardify.probability import ProbabilityField


# GENERAL SETUP
# #############


DATABASE_PROXY = DatabaseProxy()


class BaseModel(Model):
    """
    This is the abstract base class for all the models of this program. It defines the connection to a database proxy
    (a placeholder for a database), which can be bound to an actual instance of a database at runtime (when the login
    info to the database has been read from a config file possibly)

    CHANGELOG

    Added 09.06.2019
    """
    class Meta:
        database = DATABASE_PROXY


# ACTUAL MODEL CLASSES
# ####################

class User(BaseModel):
    """
    The database model to represent a user of the rewardify program

    CHANGELOG

    Added 09.06.2019
    """
    name = CharField(unique=True)
    password = PasswordField()
    gold = IntegerField()
    dust = IntegerField()

    # REWARD MANAGEMENT
    # -----------------

    def use_reward(self, reward_name: str):
        """
        Given the reward name of a type of reward in the users inventory, this method will user the reward. This means
        the reward effect will be evaluated first (if there is one) and then the instance will be deleted.

        CHANGELOG

        Added 15.06.2019

        :param reward_name:
        :return:
        """
        rewards = self.get_rewards(reward_name)
        reward = rewards[0]
        reward.use()

    def add_reward(self, reward_parameters: Dict):
        """
        Given the parameters dict for a new reward object, this method will add the specified reward to the user

        CHANGELOG

        Added 12.06.2019

        :param reward_parameters:
        :return:
        """
        # The reward parameters passed to this method are obviously still missing the "user" key, which defines to which
        # user the reward will be associated, when it is created
        reward_parameters.update({'user': self})
        reward = Reward(**reward_parameters)
        reward.save()

    def buy_reward(self, reward_parameters: Dict):
        """
        Given the parameters dict for a new Reward object, this method will buy the specified reward for the user. If
        the user does not have enough dust to buy it this method will raise a PermissionError

        CHANGELOG

        Added 12.06.2019

        :raises: PermissionError

        :param reward_parameters:
        :return:
        """
        # First we need to check if we have enough dust to buy the reward
        self.check_dust(reward_parameters['dust_cost'])

        # Now we create new pack, add it to the user and then remove the dust cost from the dust balance
        self.add_reward(reward_parameters)
        self.dust -= reward_parameters['dust_cost']

    def recycle_reward(self, reward_name: str):
        """
        Given the name of a reward type, this method will delete one of these rewards, which the user owns and add
        the recycle amount for that reward to the users dust balance.
        If the user does not own a reward of that type, LookupError will occur

        CHANGELOG

        Added 12.06.2019

        Changed 15.06.2019
        Now the recycling "operation" is not done in here, because that is none of the concern of the user object,
        instead the "recycle()" method of the Reward instance is called and the returned int amount of dust is added
        to the dust balance of the user.

        :raises: LookupError

        :param reward_name:
        :return:
        """
        rewards = self.get_rewards(reward_name)
        reward: Reward = rewards[0]
        self.dust += reward.recycle()

    def get_rewards(self, reward_name: str) -> List:
        """
        Given the name of a reward type, this method will return all reward instances of that type, which the user
        owns. If the user does not own one of this type, the method will raise a lookup error.

        CHANGELOG

        Added 12.06.2019

        :raises: LookupError

        :param reward_name:
        :return:
        """
        # Here we are getting all the rewards, that belong to this user and have the given name
        query = Reward.select(Reward, User).join(User).where(
            (User.name == self.name) &
            (Reward.name == reward_name)
        )
        if len(query) == 0:
            raise LookupError(
                "User {} does not posses rewards by the name {}".format(self.name, reward_name)
            )
        return list(query)

    def get_rewards_by_name(self) -> Dict[str, List]:
        """
        This method returns a dict, where the keys are the names of the rewards and the values are list containing the
        Reward objects from the users inventory of that reward type

        CHANGELOG

        Added 14.06.2019

        :return:
        """
        rewards_dict = defaultdict(list)
        for reward in self.rewards:
            rewards_dict[reward.name].append(reward)
        return rewards_dict

    # PACK MANAGEMENT
    # ---------------

    def add_pack(self, pack_parameters: Dict):
        """
        Given the parameters dict for a new Pack object, this method will add the specified Pack to the user

        CHANGELOG

        Added 12.06.2019

        :param pack_parameters:
        :return:
        """
        # The pack parameters passed to this method are obviously still missing the "user" key, which defines to which
        # user the pack will be associated, when it is created
        pack_parameters.update({'user': self})
        pack = Pack(**pack_parameters)
        pack.save()

    def buy_pack(self, pack_parameters: Dict):
        """
        Given the parameters dict for a new Pack object, this method will buy the pack for the user. If the user does
        not have enough gold to buy it, this method will raise a PermissionError

        CHANGELOG

        Added 12.06.2019

        :raises: PermissionError

        :param pack_parameters:
        :return:
        """
        # First we need to check if we even have enough gold to buy this.
        # In case we do not, this will raise a Permission error, otherwise do nothing
        self.check_gold(pack_parameters['gold_cost'])

        # We buy the pack by creating the pack object and assigning it to the user and then removing the gold it
        # cost from the balance
        self.add_pack(pack_parameters)
        self.gold -= pack_parameters['gold_cost']

    def open_pack(self, pack_name: str, available_rewards: List[Dict]):
        """
        Given the name of the pack type to be opened from the users inventory and a list, which will contain the
        reward parameter list for all available rewards, this method will evaluate the slot probabilities for each of
        the 5 slots and randomly pick a reward. These rewards will be added to the user and the opened pack will be
        removed.
        If the user does not own a pack of the given type, a LookupError will be risen

        CHANGELOG

        Added 12.06.2019

        :raises: LookupError

        :param pack_name:
        :param available_rewards:
        :return:
        """
        pack = self.get_packs(pack_name)[0]
        rarity_reward_map = self._available_rewards_by_rarity(available_rewards)

        for slot in pack.get_slots():
            # The "choice" method of a RarityProbability object will evaluate the probabilities and according to those
            # randomly pick and return a rarity (string)
            rarity = slot.choice()

            # When we already have a rarity, than we can get a list of all available rewards with this rarity from the
            # "rarity_reward_map". This is the second step of the random process to get the rewards. from all these
            # rewards with the same, chosen rarity one is being picked with equal probabilities
            rarity_rewards = rarity_reward_map[rarity]
            reward_parameters = np.random.choice(rarity_rewards, 1)

            # From the parameter dict, we can now add the rewards to the user directly
            self.add_reward(reward_parameters)

        # At the end we should not forget to delete the pack in question, as it was used up
        pack.delete_instance()

    def use_pack(self, pack_name: str):
        """
        Given the pack name, this method will delete one of the pack instances of the given type, which the user
        currently owns (=> using it up). If the user does not own a pack of this type. This method will raise a
        LookupError.
        NOTE!: It is important to understand, that this method does not actually perform a pack opening. To do that
        the knowledge of the configuration environment is needed, which is none of a models concern! The logic for
        actually choosing the rewards and adding them needs to be implemented in a separate function

        CHANGELOG

        Added 12.06.2019

        :raises: LookupError

        :param pack_name:
        :return:
        """
        packs = self.get_packs(pack_name)
        pack = packs[0]
        pack.delete_instance()

    def get_packs(self, pack_name: str) -> List:
        """
        Given the name of a pack type, this method will return all pack instances of that type, which the user
        owns. If the user does not own one of this type, the method will raise a lookup error.

        CHANGELOG

        Added 12.06.2019

        :raises: LookupError

        :param pack_name:
        :return:
        """
        query = Pack.select(Pack, User).join(User).where(
            (User.name == self.name) &
            (Pack.name == pack_name)
        )
        if len(query) == 0:
            raise LookupError(
                "User {} does not posse packs by the name {}".format(self.name, pack_name)
            )
        return list(query)

    def get_packs_by_name(self) -> Dict[str, List]:
        """
        This method returns a dict, where the keys are the names of different types of packs and the values are lists
        with the pack objects of these types from the users inventory

        CHANGELOG

        Added 14.06.2019

        :return:
        """
        packs_dict = defaultdict(list)
        for pack in self.packs:
            packs_dict[pack.name].append(pack)
        return packs_dict

    # CURRENCY MANAGEMENT
    # -------------------

    def check_gold(self, amount: int):
        """
        Given an amount of gold, this method checks if the user has that much gold and in case he does not it raises a
        PermissionError.

        CHANGELOG

        Added 12.06.2019

        :raises 12.06.2019

        :param amount:
        :return:
        """
        if self.gold < amount:
            raise PermissionError(
                "User {} does not have {} gold!".format(self.name, amount)
            )

    def check_dust(self, amount: int):
        """
        Given an amount of dust, this method checks if the user has that much dust and in case he does not it raises a
        PermissionError

        CHANGELOG

        Added 12.06.2019

        :raises: PermissionError

        :param amount:
        :return:
        """
        if self.dust < amount:
            raise PermissionError(
                'User {} does not habe {} dust!'.format(self.name, amount)
            )

    # HELPER FUNCTIONS
    # ----------------

    @classmethod
    def _available_rewards_by_rarity(cls, available_rewards: List[Dict]) -> Dict:
        """
        Given the list containing the parameter dicts for all available rewards, this method returns a dict, whose
        keys are the rarity indentifiers and the values are lists with all the parameter dicts of the rewards with
        that rarity

        CHANGELOG

        Added 12.06.2019

        :param available_rewards:
        :return:
        """
        mapping = {}
        for rarity in Rarity.RARITIES:
            rewards = list(filter(rarity_filter(rarity), available_rewards))
            mapping[rarity] = rewards

        return mapping


class Reward(BaseModel):
    """
    The database model to represent a reward, which a user has obtained

    CHANGELOG

    Added 09.06.2019

    Changed 15.06.2019
    Added the "recycle" method, which will delete the instance and return the amount of dust gained from the recycling.
    Added the "use" method, which will use up the reward.
    Added the "effect" text field for the new version. This field can optionally contain a special syntax, which
    specified, what effect the reward has on the rewardify system, when it is being used. As of right now two effects
    are being supported: Granting gold and dust to the user upon opening.
    """

    # DATABASE FIELDS
    # ---------------

    name = CharField()
    slug = CharField()
    description = TextField()
    dust_cost = IntegerField()
    dust_recycle = IntegerField()
    date_obtained = DateTimeField()
    rarity = RarityField()
    effect = TextField()

    user = ForeignKeyField(User, backref='rewards')

    # CONSTANTS
    # ---------

    # These regular expressions will search for any occurance of the substrings "gold(...)" and "dust(...)".
    # The result of the regular expressions will also directly exctract the substring from within the parentheses,
    # which specify the amount that has been granted by this effect
    GOLD_EFFECT_REGEX = 'gold\((.+?)\)'
    DUST_EFFECT_REGEX = 'dust\((.+?)\)'

    # INSTANCE METHODS
    # ----------------

    def use(self):
        """
        This method will use up the reward. First any possible effect of the reward will be evaluated and then the
        instance is deleted

        CHANGELOG

        Added 15.06.2019

        :return:
        """
        self.evaluate_effect()
        self.delete_instance()

    def recycle(self):
        """
        This method will recycle the reward, which means, that the instance will be deleted, but the method will return
        the integer amount of dust, that has been gained from recycling the reward

        CHANGELOG

        Added 15.06.2019

        :return:
        """
        dust_recycle = self.dust_recycle
        self.delete_instance()
        return dust_recycle

    def evaluate_effect(self):
        """
        This method evaluates the effect given in the "effect" string of the reward instance. The effect string can
        (it is an optional property of a reward) contain special syntax to describe, what kind of effect the reward has
        on the rewardify system, when it is being used. (Such as adding gold to the users balance)

        CHANGELOG

        Added 15.06.2019

        :return:
        """
        self.evaluate_dust_effect()
        self.evaluate_gold_effect()
        self.user.save()

    # EVALUATION OF THE EFFECTS
    # -------------------------

    def evaluate_gold_effect(self):
        """
        This method will evaluate the "effect" string of the reward instance to see if it contains the syntax
        "gold(...)" for a gold effect and if it does the effect is being applied by adding as much gold as specified
        in the parentheses to the users gold.
        Raises a value error if the argument within the parentheses is not an integer.

        CHANGELOG

        Added 15.06.2019

        :raise: ValueError

        :return:
        """
        try:
            effect = str(self.effect)
            result = re.search(self.GOLD_EFFECT_REGEX, effect)
            substring = result.group(1)
            amount = int(substring)
            self.user.gold += amount
        # An attribute error will occur, when the regular expression has not found anything and it is attempted to call
        # the "group(1)" method on a none type object.
        # In this case there just is no effect to be applied
        except AttributeError:
            return
        # A value error will occur, when there was this expression in the effect string, but the substring in between
        # the parentheses is not the string of an integer.
        # This is clearly a problem and thus the exception is being risen again
        except ValueError:
            raise ValueError('The substring "{}" can not be the argument of a GOLD EFFECT!'.format(substring))

    def evaluate_dust_effect(self):
        """
        This method will evaluate the "effect" string of the reward instance to see if it contains the syntax
        "dust(...)" for a dust effect and if it does, the effect is being applied by adding as much dust as specified
        in the parentheses to the users dust account.

        Raises a value error if the argument within the parentheses is not an integer.

        CHANGELOG

        Added 15.06.2019

        :raise: ValueError

        :return:
        """
        try:
            effect = str(self.effect)
            result = re.search(self.DUST_EFFECT_REGEX, effect)
            substring = result.group(1)
            amount = int(substring)
            self.user.dust += amount
        # An attribute error will occur, when the regular expression has not found anything and it is attempted to call
        # the "group(1)" method on a none type object.
        # In this case there just is no effect to be applied
        except AttributeError:
            return
        # A value error will occur, when there was this expression in the effect string, but the substring in between
        # the parentheses is not the string of an integer.
        # This is clearly a problem and thus the exception is being risen again
        except ValueError:
            raise ValueError('The substring "{}" can not be the argument of a DUST EFFECT!'.format(substring))


class Pack(BaseModel):
    """
    The database model to represent a pack, which a user has obtained

    CHANGELOG

    Added 09.06.2019
    """
    SLOT_INDICES = [1, 2, 3, 4, 5]

    name = CharField()
    slug = CharField()
    description = TextField()
    gold_cost = IntegerField()
    date_obtained = DateTimeField()
    slot1 = ProbabilityField()
    slot2 = ProbabilityField()
    slot3 = ProbabilityField()
    slot4 = ProbabilityField()
    slot5 = ProbabilityField()

    user = ForeignKeyField(User, backref='packs')

    def get_slots(self):
        """
        This will return a list which contains the RarityProbability objects for the 5 slots of this pack instance.

        CHANGELOG

        Added 12.06.2019

        :return:
        """
        slots = []
        for index in self.SLOT_INDICES:
            # This will basically return the field object of this very Pack instance for the fields in ascending
            # order
            slot = getattr(self, f'slot{index}')
            slots.append(slot)

        return slots




# Standard library
import platform
import os
import shutil

from unittest import TestCase

from typing import Dict, List

# Third party
from peewee import SqliteDatabase

from jinja2 import Environment, FileSystemLoader

# Local imports
from rewardify.models import DATABASE_PROXY
from rewardify.models import User, Pack, Reward

from rewardify.env import EnvironmentConfig


PATH = os.path.dirname(os.path.abspath(__file__))


# ###########
# FOR TESTING
# ###########


# THE MIXINS
# ----------


class ConfigTestMixin:
    """
    This class implements TestCase behaviour to create a new config folder and configure the
    ConfigEnvironment to use this correctly on class setup.
    It also implements various additional helper methods, that can be used to create custom config
    environments for each test method.

    CHANGELOG

    Added 10.06.2019
    """
    # OVERWRITING THE CONFIG PATH OF THE ENV
    # --------------------------------------

    # According to the operation system, on which the program is being run, a different path has to be used for the
    # config folder. This mapping assigns the path to be used for testing to each operating system returned by
    # "platform.system"
    PLATFORM_TEST_CONFIG_PATHS = {
        'Linux': '/tmp/rewardify',
        'Darwin': '',
        'Windows': '',
    }

    # This field will store the path which was originally set within the ConfigEnvironment, it will be used to reset
    # "everything back to how we found it" at tear down
    ORIGINAL_FOLDER_PATH = ''
    # This field will store the actual folder path to the config folder after the decision based on the operating
    # system has been made.
    FOLDER_PATH = ''

    # FOR TEMPLATING THE CONFIG FILES
    # -------------------------------

    TEMPLATE_FOLDER_PATH = os.path.join(PATH, 'templates', 'test')
    TEMPLATE_LOADER = FileSystemLoader(searchpath=TEMPLATE_FOLDER_PATH)
    TEMPLATE_ENVIRONMENT = Environment(loader=TEMPLATE_LOADER)

    CONFIG_TEMPLATE_NAME = 'config.jinja2'
    CONFIG_FILE_NAME = 'config.py'

    # ################
    # INSTANCE METHODS
    # ################

    def setUp(self):
        """
        This method is called before the execution of every test method. It calls the "clean" method, which deletes all
        the files from the config folder, so the test method starts out with a clean slate.

        CHANGELOG

        Added 10.06.2019

        :return:
        """
        self.clean()

    # #############
    # CLASS METHODS
    # #############

    @classmethod
    def setUpClass(cls):
        """
        This method is called before any of the test methods of this class are being executed. It creates the config
        folder to be used for testing and configures the ConfigEnvironment singleton to use this path instead of its
        default install location.

        CHANGELOG

        Added 10.06.2019

        :return:
        """
        # We find out, which path to use for the testing by querying the current operating system
        platform_string = platform.system()
        cls.FOLDER_PATH = cls.PLATFORM_TEST_CONFIG_PATHS[platform_string]

        # Now we save the original path of the ConfigEnvironment and replace it with the testing path
        environment_config = EnvironmentConfig.instance()  # type: EnvironmentConfig
        cls.ORIGINAL_FOLDER_PATH = environment_config.folder_path
        environment_config.set_folder_path(cls.FOLDER_PATH)

        # Now we only need to create the config folder
        cls.create_config_folder()

    @classmethod
    def create_config_folder(cls):
        """
        This method creates a new folder with the path "cls.FOLDER_PATH", if it does not already exist

        CHANGELOG

        Added 10.06.2019

        :return:
        """
        if not os.path.exists(cls.FOLDER_PATH):
            os.makedirs(cls.FOLDER_PATH)

    @classmethod
    def tearDownClass(cls):
        # Now we set back the original folder path
        environment_config: EnvironmentConfig = EnvironmentConfig.instance()
        environment_config.set_folder_path(cls.ORIGINAL_FOLDER_PATH)

    @classmethod
    def create_config(cls,
                      imports: List,
                      database_dict: Dict,
                      backend,
                      rewards: List,
                      packs: List,
                      plugin_code: str):
        """
        Given a list with python code strings for import statements "imports", a dict defining the database access
        "database_dict", the string containing the class name of the backend class to be used "backend", a list
        defining the available "rewards" and "packs" and a python code string for additional variable definitions in
        the config file, that might be needed by plugins.
        This method uses all of these to create a new "config.py" file in the "cls.FOLDER_PATH" based on
        the template for the config file

        CHANGELOG

        Added 10.06.2019

        :param imports:
        :param database_dict:
        :param backend:
        :param rewards:
        :param packs:
        :param plugin_code:
        :return:
        """
        context = {
            'imports':      imports,
            'db':           database_dict,
            'backend':      backend,
            'rewards':      rewards,
            'packs':        packs,
            'plugin':       plugin_code
        }
        config_file_path = os.path.join(cls.FOLDER_PATH, cls.CONFIG_FILE_NAME)
        cls.file_from_template(
            config_file_path,
            cls.CONFIG_TEMPLATE_NAME,
            context
        )

        # When a new config file is created, we obviously want to see these changes in the config object as well,
        # this is why we are reloading the config object with the new config file
        environment_config: EnvironmentConfig = EnvironmentConfig.instance()
        environment_config.load()

    # HELPER METHODS
    # --------------

    @classmethod
    def clean(cls):
        """
        This method deletes all the files within the "cls.FOLDER_PATH"

        CHANGELOG

        Added 10.06.2019

        :return:
        """
        for the_file in os.listdir(cls.FOLDER_PATH):
            file_path = os.path.join(cls.FOLDER_PATH, the_file)
            try:
                if os.path.isfile(file_path):
                    os.unlink(file_path)
                # elif os.path.isdir(file_path): shutil.rmtree(file_path)
            except Exception as e:
                print(e)

    @classmethod
    def get_config_file_path(cls):
        """
        This method returns the file path to the "config.py" file within the testing config folder

        CHANGELOG

        Added 10.06.2019

        :return:
        """
        return os.path.join(cls.FOLDER_PATH, cls.CONFIG_FILE_NAME)

    @classmethod
    def file_from_template(cls, file_path: str, template_name: str, context: Dict):
        """
        Given the string path of a file to be written to, the name of the template (within the defined template folder)
        and a dict defining the context of for the jinja template, this method will render the template with the
        context and save the result into the given file path.

        CHANGELOG

        Added 10.06.2019

        :param file_path:
        :param template_name:
        :param context:
        :return:
        """
        # First we render the template content into a plain string and then we save the string into the
        # file (overwriting all the content)
        template = cls.TEMPLATE_ENVIRONMENT.get_template(template_name)
        content = template.render(**context)

        with open(file_path, mode='w+') as file:
            file.write(content)


class DBTestMixin:
    """
        This is a abstract base class, which adds additional setUp and tearDown steps to create a fresh
        in memory database before each test method is being executed, which enables the work with the model classes.

        CHANGELOG

        Added 09.06.2019
        """
    TEST_DATABASE = SqliteDatabase(':memory:')
    MODELS = [
        User,
        Pack,
        Reward
    ]

    def setUp(self):
        """
        This method is being called before each test method is executed.
        It creates a new in-memory database according to the models configured in the "models.py" file of
        this project

        CHANGELOG

        Added 09.06.2019

        :return:
        """
        # Here we bind the database proxy (placeholder) for which we have defined the models
        # to the test database
        DATABASE_PROXY.initialize(self.TEST_DATABASE)
        self.TEST_DATABASE.connect()
        self.TEST_DATABASE.create_tables(self.MODELS)

    def tearDown(self):
        """
        This method is being called after each test method was executed.
        It closes the database that was created for the test method

        CHANGELOG

        Added 09.06.2019

        :return:
        """
        self.TEST_DATABASE.drop_tables(self.MODELS)
        self.TEST_DATABASE.close()


# ACTUAL ABSTRACT TEST CASE BASE CLASSES
# --------------------------------------


class ConfigTestCase(ConfigTestMixin, TestCase):

    pass


class DBTestCase(DBTestMixin, TestCase):

    pass


class RewardifyTestCase(ConfigTestMixin, DBTestMixin, TestCase):

    def setUp(self):
        ConfigTestMixin.setUp(self)
        DBTestMixin.setUp(self)




# standard library
import datetime

from typing import Dict, List

# local imports
from rewardify._util_test import RewardifyTestCase

from rewardify.backends import MockBackend

from rewardify.models import User, Pack, Reward

from rewardify.main import Rewardify
from rewardify.main import available_rewards, available_rewards_by_rarity
from rewardify.main import open_pack


class TestMain(RewardifyTestCase):

    # SAMPLE VALUES FOR THE ENVIRONMENT CONFIG
    # ----------------------------------------

    SAMPLE_IMPORTS = [
        'from rewardify.backends import AbstractBackend'
    ]

    SAMPLE_DATABASE_DICT = {
        'engine': 'sqlite',
        'host': ':memory:',
        'database': '',
        'user': '',
        'password': '',
        'port': 0
    }

    SAMPLE_PACKS = [
        {
            'name':             'Sample Pack',
            'cost':             100,
            'description':      'for testing',
            'slot1':            [1, 0, 0, 0],
            'slot2':            [1, 0, 0, 0],
            'slot3':            [1, 0, 0, 0],
            'slot4':            [1, 0, 0, 0],
            'slot5':            [1, 0, 0, 0],
        }
    ]

    SAMPLE_REWARDS = [
        {
            'name':             'Sample Reward',
            'description':      'for testing',
            'rarity':           'uncommon',
            'cost':             100,
            'recycle':          100
        }
    ]

    SAMPLE_BACKEND = 'MockBackend'

    SAMPLE_PLUGIN_CODE = 'MOCK_BACKEND_USERS = ["Jonas"]'

    # SAMPLE DATABASE VALUES
    # ----------------------

    SAMPLE_USER_PARAMETERS = {
        'name':     'Jonas',
        'password':     'secret',
        'gold':         0,
        'dust':         0
    }

    # TESTING BUSINESS LOGIC FUNCTIONS
    # --------------------------------

    def test_available_rewards(self):
        # For the available rewards we need to config environment
        self.setup_sample()

        rewards = available_rewards()
        filtered_rewards = list(filter(lambda d: d['name'] == 'Sample Reward', rewards))
        self.assertEqual(len(filtered_rewards), 1)

    def test_available_rewards_by_rarity(self):
        self.setup_sample()

        rarity_reward_map = available_rewards_by_rarity()
        self.assertEqual(len(rarity_reward_map['uncommon']), 1)
        self.assertEqual(len(rarity_reward_map['legendary']), 0)

    def test_open_pack(self):
        self.setup_sample()
        self.create_sample_user()
        user = self.get_sample_user()

        pack_parameters = {
            'name':             'Sample Pack',
            'slug':             'sample_pack',
            'description':      'testing',
            'gold_cost':        100,
            'date_obtained':    datetime.datetime.now(),
            'slot1':            [1, 0, 0, 0],
            'slot2':            [1, 0, 0, 0],
            'slot3':            [1, 0, 0, 0],
            'slot4':            [1, 0, 0, 0],
            'slot5':            [1, 0, 0, 0],
        }

        # First we need to give the user one pack
        user.add_pack(pack_parameters)

        open_pack(
            self.SAMPLE_USER_PARAMETERS['name'],
            'Sample Pack'
        )

        self.assertEqual(len(user.packs), 0)
        self.assertEqual(len(user.rewards), 5)

    # TESTING FACADE OBJECT
    # ---------------------

    def test_test_case_working(self):
        self.setup_sample()

        # Now if everything was setup correctly the config instance from within the facade object should contain all
        # the sample values
        facade = Rewardify.instance()
        self.assertTrue(hasattr(facade.CONFIG, "BACKEND"))
        self.assertEqual(facade.CONFIG.BACKEND, MockBackend)

    def test_creating_new_user(self):
        facade: Rewardify = Rewardify.instance()

        # Creating the user
        username = 'Nigel'
        password = 'snobby'
        facade.create_user(username, password)

        # Querying for the user to check if it exists
        query = User.select().where(User.name == username)
        self.assertTrue(len(query) >= 1)
        user = query[0]
        self.assertTrue(user.password.check(password))

    def test_checking_password_to_username(self):
        # Setting up the environment and Inserting a new user into the database
        self.setup_sample()
        self.create_sample_user()

        facade: Rewardify = Rewardify.instance()
        facade.user_check_password(
            self.SAMPLE_USER_PARAMETERS['name'],
            self.SAMPLE_USER_PARAMETERS['password']
        )

    def test_getting_user_model_from_facade(self):
        # Setting up the environment and Inserting a new user into the database
        self.setup_sample()
        self.create_sample_user()

        facade: Rewardify = Rewardify.instance()
        user = facade.get_user(self.SAMPLE_USER_PARAMETERS['name'])
        self.assertIsInstance(user, User)

    def test_available_packs(self):
        self.setup_sample()

        facade: Rewardify = Rewardify.instance()
        available_packs: List = facade.available_packs()
        self.assertTrue(len(available_packs), 2)

    def test_backend_update(self):
        self.setup_sample()
        self.create_sample_user()
        user = self.get_sample_user()
        self.assertEqual(user.gold, 0)

        facade: Rewardify = Rewardify.instance()
        facade.backend_update()

        # The mock backend gives every user specified in MOCK_BACKEND_USERS exactly 100 gold on every call of the
        # update, which means, that after the call the user should have been granted 100 gold from the mock backend.
        user = self.get_sample_user()
        self.assertEqual(user.gold, 100)

    def test_user_exists(self):
        self.setup_sample()
        facade: Rewardify = Rewardify.instance()

        # Of course from the get go the user is not in the database, which means the exists call should return false
        user_exists = facade.exists_user(self.SAMPLE_USER_PARAMETERS['name'])
        self.assertFalse(user_exists)

        # After we have inserted the user though the exists call should return true
        self.create_sample_user()
        user_exists = facade.exists_user(self.SAMPLE_USER_PARAMETERS['name'])
        self.assertTrue(user_exists)

    # HELPER METHODS
    # --------------

    def setup_sample(self):
        # Setting up the config file with the sample values
        self.create_config(
            self.SAMPLE_IMPORTS,
            self.SAMPLE_DATABASE_DICT,
            self.SAMPLE_BACKEND,
            self.SAMPLE_REWARDS,
            self.SAMPLE_PACKS,
            self.SAMPLE_PLUGIN_CODE
        )

    def create_sample_user(self):
        user = User(**self.SAMPLE_USER_PARAMETERS)
        user.save()

    def get_sample_user(self) -> User:
        user = User.get(User.name == self.SAMPLE_USER_PARAMETERS['name'])
        return user



# Standard library
from unittest import TestCase

# Third party
from peewee import SqliteDatabase
from peewee import Model
from peewee import CharField

# Local imports
from rewardify.rarity import Rarity
from rewardify.rarity import RarityField


class TestRarity(TestCase):

    # ACTUAL TEST CASES
    # -----------------

    def test_construction_with_string_works(self):
        rarity_string = 'legendary'
        rarity_int = 4
        rarity = Rarity(rarity_string)

        self.assertEqual(str(rarity), rarity_string)
        self.assertEqual(int(rarity), rarity_int)

    def test_construction_with_int_works(self):
        rarity_string = 'legendary'
        rarity_int = 4
        rarity = Rarity(rarity_int)

        self.assertEqual(str(rarity), rarity_string)
        self.assertEqual(int(rarity), rarity_int)

    def test_comparison_works(self):
        rarity_high = Rarity(4)
        rarity_low = Rarity(1)

        self.assertTrue(rarity_high > rarity_low)
        self.assertTrue(rarity_low < rarity_high)
        self.assertTrue(rarity_high == rarity_high)


# TESTING THE RARITY FIELD
# ------------------------

class RarityModel(Model):

    name = CharField()
    rarity = RarityField()


class TestRarityField(TestCase):

    TEST_DATABASE = SqliteDatabase(':memory:')
    MODELS = [RarityModel]

    SAMPLE_MODEL_NAME = 'sample'

    # SETTING UP THE TEST DATABASE
    # ----------------------------

    def setUp(self):
        self.TEST_DATABASE.bind(self.MODELS, bind_refs=False, bind_backrefs=False)

        self.TEST_DATABASE.connect()
        self.TEST_DATABASE.create_tables(self.MODELS)

    def tearDown(self):
        self.TEST_DATABASE.drop_tables(self.MODELS)
        self.TEST_DATABASE.close()

    # THE ACTUAL TEST CASES
    # ---------------------

    def test_rarity_field_insert_string_works(self):
        # Inserting a model into the database
        model = RarityModel(
            name=self.SAMPLE_MODEL_NAME,
            rarity='legendary'
        )
        model.save()

        # Loading it again
        model = RarityModel.get(RarityModel.name == self.SAMPLE_MODEL_NAME)
        self.assertIsInstance(model.rarity, Rarity)
        self.assertEqual(model.rarity, Rarity.LEGENDARY)

    def test_rarity_field_insert_int_works(self):
        model = RarityModel(
            name=self.SAMPLE_MODEL_NAME,
            rarity=4
        )
        model.save()

        model = self.get_sample_model()
        self.assertIsInstance(model.rarity, Rarity)
        self.assertEqual(model.rarity, Rarity.LEGENDARY)

    # HELPER METHODS
    # --------------

    def get_sample_model(self):
        return RarityModel.get(RarityModel.name == self.SAMPLE_MODEL_NAME)



# Standard library
import os
import shutil
import platform

from unittest import TestCase

# Third party
from peewee import SqliteDatabase

from jinja2 import FileSystemLoader, Environment

# Local imports
from rewardify.env import DatabaseConfig
from rewardify.env import EnvironmentConfig
from rewardify.env import EnvironmentInstaller

from rewardify.models import User
from rewardify.models import DATABASE_PROXY

from rewardify._util_test import ConfigTestCase


FOLDER_PATH = os.path.dirname(os.path.abspath(__file__))


class TestDatabaseConfig(TestCase):

    # CLASS VARIABLES
    # ---------------

    SAMPLE_DATABASE_DICT = {
        'engine':       'sqlite',
        'host':         ':memory:'
    }

    # ACTUAL TEST CASES
    # -----------------

    def test_object_creation(self):
        database_config = DatabaseConfig(
            'sqlite',
            ':memory:',
            '',
            '',
            '',
            0
        )
        self.assertEqual(database_config.engine, 'sqlite')

    def test_object_creation_from_dict(self):
        database_dict = {
            'engine':       'postgres',
            'user':         'Jonas',
            'host':         'localhost'
        }
        database_config = DatabaseConfig.from_dict(database_dict)

        self.assertIsInstance(database_config, DatabaseConfig)
        self.assertEqual(database_config.user, 'Jonas')

    def test_checking_engine_string_validity(self):
        database_config = DatabaseConfig.from_dict({'engine': 'sqlite'})
        self.assertEqual(database_config.engine, 'sqlite')

        self.assertRaises(
            ValueError,
            DatabaseConfig.from_dict,
            {'engine':  'yolo-db'}
        )

    def test_database_object_creation_sqlite(self):
        database_config = self.get_sample_object()
        database_config.init()

        self.assertIsInstance(database_config.get(), SqliteDatabase)

    def test_sqlite_database(self):
        database_config = self.get_sample_object()
        database_config.init()
        DATABASE_PROXY.initialize(database_config.get())
        database_config.connect()
        database_config.create_tables()

        # Creating a user and saving it just to see if it works
        user = User(
            name="Jonas",
            password="secret",
            dust=0,
            gold=0,
        )
        user.save()

    # HELPER METHODS
    # --------------

    def get_sample_object(self):
        return DatabaseConfig.from_dict(self.SAMPLE_DATABASE_DICT)


class TestEnvironmentConfig(ConfigTestCase):

    SAMPLE_IMPORTS = [
        'from rewardify.backends import AbstractBackend'
    ]

    SAMPLE_DATABASE_DICT = {
        'engine':       'sqlite',
        'host':         ':memory:',
        'database':     '',
        'user':         '',
        'password':     '',
        'port':         0
    }

    SAMPLE_PACKS = [

    ]

    SAMPLE_REWARDS = [

    ]

    SAMPLE_BACKEND = 'MockBackend'

    SAMPLE_PLUGIN_CODE = 'MOCK_BACKEND_USERS = ["Jonas"]'

    # ACTUAL TEST METHODS
    # -------------------

    def test_config_test_case_working(self):
        self.assertTrue(os.path.exists(self.FOLDER_PATH))

    def test_create_config(self):
        self.setup_sample()
        self.assertTrue(os.path.exists(self.get_config_file_path()))

    def test_load_config_file(self):
        self.setup_sample()

        environment_config: EnvironmentConfig = EnvironmentConfig.instance()
        environment_config.load_config()

        self.assertTrue(hasattr(environment_config, 'DATABASE'))
        self.assertTrue(hasattr(environment_config, 'BACKEND'))

    # HELPER METHODS
    # --------------

    def setup_sample(self):
        # Setting up the config-py file
        self.create_config(
            self.SAMPLE_IMPORTS,
            self.SAMPLE_DATABASE_DICT,
            self.SAMPLE_BACKEND,
            self.SAMPLE_REWARDS,
            self.SAMPLE_PACKS,
            self.SAMPLE_PLUGIN_CODE
        )


class TestEnvironmentInstaller(TestCase):

    FOLDER_PATH = '/tmp/rewardify'

    # SETTING UP
    # ----------

    def setUp(self):
        # Before every test, we need to clean up the environment
        installation_folder_path = self.get_installation_path()
        if os.path.exists(installation_folder_path):
            shutil.rmtree(installation_folder_path)

    # ACTUAL TESTS
    # ------------

    def test_init_working(self):
        installer = EnvironmentInstaller(self.FOLDER_PATH)
        self.assertEqual(installer.path, self.FOLDER_PATH)

    def test_get_file_paths(self):
        installer = EnvironmentInstaller(self.FOLDER_PATH)
        folder_path = os.path.join(self.FOLDER_PATH, installer.PLATFORM_CONFIG_FOLDER_NAME[platform.system()])
        self.assertEqual(
            installer.get_config_folder_path(),
            folder_path
        )
        self.assertEqual(
            installer.get_config_file_path(),
            os.path.join(folder_path, installer.CONFIG_FILE_NAME)
        )
        self.assertEqual(
            installer.get_database_file_path(),
            os.path.join(folder_path, installer.DATABASE_FILE_NAME)
        )

    def test_config_file_created(self):
        installer = EnvironmentInstaller(self.FOLDER_PATH)
        installer.create_config_folder()
        installer.create_config_file()

        self.assertTrue(os.path.exists(installer.get_config_file_path()))

    def test_database_file_created(self):
        installer = EnvironmentInstaller(self.FOLDER_PATH)
        installer.create_config_folder()
        installer.create_database_file()

        self.assertTrue(os.path.exists(installer.get_database_file_path()))

    def test_database_creation_successful(self):
        installer = EnvironmentInstaller(self.FOLDER_PATH)
        installer.create_config_folder()
        installer.create_database_file()

        self.assertTrue(os.path.exists(installer.get_database_file_path()))
        # The Database should be connected to the models now, which one should be able to create a new model and
        # insert it into the database
        user = User(
            name="Jonas",
            password="Secret",
            gold=0,
            dust=0
        )
        user.save()
        self.assertIsInstance(user, User)

    def test_config_creation_successful(self):
        installer = EnvironmentInstaller(self.FOLDER_PATH)
        installer.create_config_folder()
        installer.create_config_file()

        self.assertTrue(os.path.exists(installer.get_config_file_path()))
        with open(installer.get_config_file_path(), mode="r") as file:
            content = file.read()
            self.assertTrue("DATABASE" in content)
            self.assertTrue("BACKEND" in content)
            self.assertTrue("MOCK_BACKEND_USERS" in content)

    # HELPER METHODS
    # --------------

    def get_installation_path(self):
        return os.path.join(self.FOLDER_PATH, EnvironmentInstaller.PLATFORM_CONFIG_FOLDER_NAME[platform.system()])



# Standard library
from typing import List, Dict

# third party
import numpy as np

# Local imports
from rewardify._util import Singleton

from rewardify.env import EnvironmentConfig

from rewardify.models import User, Pack, Reward

from rewardify.rarity import Rarity

from rewardify.adapters import RewardParametersAdapter, PackParametersAdapter

from rewardify.backends.base import AbstractBackend


# #######################
# BUSINESS LOGIC WRAPPERS
# #######################


def rarity_filter(rarity: str):
    def __callable(reward: dict):
        return reward['rarity'] == rarity
    return __callable


def available_packs() -> List[Dict]:
    """
    Returns a list of dicts, where each dict describes on of the available packs, defined in the config file:
    The dicts contain the following keys:
    name: String name of the pack
    cost: Integer gold cost
    description: string description
    1: A list with four float values describing the rarity probability distribution of that slot
    2: A list with four float values describing the rarity probability distribution of that slot
    3: A list with four float values describing the rarity probability distribution of that slot
    4: A list with four float values describing the rarity probability distribution of that slot
    5: A list with four float values describing the rarity probability distribution of that slot

    CHANGELOG

    Added 14.06.2019

    :return:
    """
    config = EnvironmentConfig.instance()
    packs_list = []
    for name, parameters in config.PACKS.items():
        packs_list.append({**parameters, **{'name': name}})
    packs_list.sort(key=lambda x: x['cost'], reverse=True)
    return packs_list


def available_rewards() -> List[Dict]:
    """
    Returns a list with dicts, each dict describing an available reward, which was defined in the backend
    The dicts contain the following keys:
    - name: String name of the reward
    - cost: Integer amount of dust, which the reward costs
    - recycle: Integer amount of dust one gets, when the reward is recycled
    - description: String description of the reward
    - rarity: The string identified for the rarity of the reward can be 'common', 'uncommon', 'rare' and
    'legendary'

    CHANGELOG

    Added 12.06.2019

    :return:
    """
    config = EnvironmentConfig.instance()
    rewards_list = []
    for name, parameters in config.REWARDS.items():
        rewards_list.append({**parameters, **{'name': name}})
    return rewards_list


def available_rewards_by_rarity():
    """
    Returns a dict, whose keys are the string rarity identifiers and the items are lists containing dicts describing
    the available rewards of that rarity, that were defined in the config file.
    The dicts contain the following keys:
    - name: String name of the reward
    - cost: Integer amount of dust, which the reward costs
    - recycle: Integer amount of dust one gets, when the reward is recycled
    - description: String description of the reward
    - rarity: The string identified for the rarity of the reward can be 'common', 'uncommon', 'rare' and
    'legendary'
    The list are sorted by the dust cost of the rewards in descending order.

    CHANGELOG

    Added 12.06.2019

    Changed 14.06.2019
    Added a sorting statement for the lists, so that they will be sorted by the their cost

    :return:
    """
    rewards = available_rewards()
    reward_dict = {}
    for rarity in Rarity.RARITIES:
        filtered_rewards = list(filter(rarity_filter(rarity), rewards))

        # 14.06.2019
        # As this function might be used to display the rewards to the user on the front end it will be useful to
        # have the entries be sorted by the cost
        filtered_rewards.sort(key=lambda x: x['cost'], reverse=True)

        reward_dict[rarity] = filtered_rewards

    return reward_dict


def open_pack(username: str, pack_name: str):
    """
    Given a user by its username and the name of a pack type, which the user owns, this function will open that pack
    for the user. This means it will evaluate the slot probabilities for all 5 slots of the pack. Based on these
    probabilities 5 rewards are randomly chosen from the set of available rewards. These rewards will be added to the
    user. The pack will be removed, as it has been used up.

    CHANGELOG

    Added 12.06.2019

    :param username:
    :param pack_name:
    :return:
    """
    user = User.get(User.name == username)
    pack = user.get_packs(pack_name)[0]

    rarity_reward_map = available_rewards_by_rarity()

    for i in range(1, 6, 1):
        slot = getattr(pack, f'slot{i}')
        rarity = slot.choice()
        reward = np.random.choice(rarity_reward_map[rarity], 1)[0]
        parameters_adapter = RewardParametersAdapter(reward['name'], reward)
        user.add_reward(parameters_adapter.parameters())

    user.use_pack(pack_name)
    user.save()

# ###############
# THE MAIN FACADE
# ###############


@Singleton
class Rewardify:
    """


    Note, 12.06.2019
    Ok I am making a design choice with this facade now. My problem was choosing between if I want to use the username
    as the parameter to all the methods of this class or do I want to use User objects. Because using just the name
    to identify the user is obviously more SOC, because whoever uses the facade doesnt have to directly interact with
    the model layer, for simple operations at least. But the user object has to be queried every time, which is more
    inefficient. I am doing it anyways, because efficiency wont be an issue with a small scale program like this
    anyways, plus one could implement caching later on if it does get a problem. So ill be acting by the theorem
    "dont overenginieer from the beginning"

    CHANGELOG

    Added 10.06.2019

    Changed 15.06.2019
    Added the method "exists_user" which returns the boolean value of a user of that name exists in the database.
    Added multiple methods for dealing with rewards including the buying, using and recycling

    """
    CONFIG: EnvironmentConfig = EnvironmentConfig.instance()

    def __init__(self):
        pass

    # ENVIRONMENTAL INFORMATION
    # -------------------------

    @classmethod
    def available_rewards_by_rarity(cls) -> Dict[str, List]:
        """
        Returns a dict, whose keys are the string rarity identifiers and the items are lists containing dicts describing
        the available rewards of that rarity, that were defined in the config file.
        The dicts contain the following keys:
        - name: String name of the reward
        - cost: Integer amount of dust, which the reward costs
        - recycle: Integer amount of dust one gets, when the reward is recycled
        - description: String description of the reward
        - rarity: The string identified for the rarity of the reward can be 'common', 'uncommon', 'rare' and
        'legendary'
        The list are sorted by the dust cost of the rewards in descending order.

        CHANGELOG

        Added 12.06.2019

        :return:
        """
        return available_rewards_by_rarity()

    @classmethod
    def available_rewards(cls) -> List[Dict]:
        """
        Returns a list with dicts, each dict describing an available reward, which was defined in the backend
        The dicts contain the following keys:
        - name: String name of the reward
        - cost: Integer amount of dust, which the reward costs
        - recycle: Integer amount of dust one gets, when the reward is recycled
        - description: String description of the reward
        - rarity: The string identified for the rarity of the reward can be 'common', 'uncommon', 'rare' and
        'legendary'

        CHANGELOG

        Added 12.06.2019

        :return:
        """
        return available_rewards()

    @classmethod
    def available_packs(cls) -> List[Dict]:
        """
        Returns a list of dicts, where each dict describes on of the available packs, defined in the config file:
        The dicts contain the following keys:
        name: String name of the pack
        cost: Integer gold cost
        description: string description
        1: A list with four float values describing the rarity probability distribution of that slot
        2: A list with four float values describing the rarity probability distribution of that slot
        3: A list with four float values describing the rarity probability distribution of that slot
        4: A list with four float values describing the rarity probability distribution of that slot
        5: A list with four float values describing the rarity probability distribution of that slot

        CHANGELOG

        Added 14.06.2019

        :return:
        """
        return available_packs()

    # USER MANAGEMENT
    # ---------------

    def user_check_password(self, username: str, password: str) -> bool:
        """
        Returns true, if the given password is the password of the user with the given username and false otherwise

        CHANGELOG

        Added 10.06.2019

        :param username:
        :param password:
        :return:
        """
        user = self.get_user(username)
        # The "password" field of a user object is a special "PasswordHash" object, which offers the method check to
        # perform a password check. A direct comparison wont work, since only a hash is saved in the database
        return user.password.check(password)

    def create_user(self, username: str, password: str):
        """
        Creates a new user with the given username and password. Initializes gold and dust with 0

        CHANGELOG

        Added 10.06.2019

        :param username:
        :param password:
        :return:
        """
        user = User(
            name=username,
            password=password,
            gold=0,
            dust=0
        )
        user.save()

    def exists_user(self, username: str):
        """
        Returns the boolean value of whether or not a user with the given username exists

        CHANGELOG

        Added 15.06.2019

        :param username:
        :return:
        """
        query = User.select().where(User.name == username)
        # Of course, when there is a result to the above query, a User object has been found and thus, the user exists
        return len(query) != 0

    # DATABASE MODELS
    # ---------------

    def get_user(self, username: str) -> User:
        """
        Returns the User object for the user with the given username

        CHANGELOG

        Added 10.06.2019

        :param username:
        :return:
        """
        query = User.select(User).where(User.name == username)
        user = query[0]
        return user

    # USER INFORMATION
    # ----------------

    # USER PACK MANAGEMENT
    # --------------------

    def user_open_pack(self, username: str, packname: str):
        """
        Opens a pack of the given packname type from the given users inventory.
        This means, that according to the packs probability distribution 5 new rewards will be added to the users
        inventory, while the pack is being removed.
        If the user does not posses a pack of that type, this method will raise a LookupError.

        CHANGELOG

        Added 12.06.2019

        :raise: LookupError

        :param username:
        :param packname:
        :return:
        """
        user = self.get_user(username)

        # The method "open_pack" of the user model expects a list of all the parameter dicts for all the available
        # rewards, so we have to create that first
        reward_parameters = self.available_rewards_parameters()

        user.open_pack(packname, reward_parameters)

    def user_buy_pack(self, username: str, packname: str):
        """
        Buys a pack of the type with the given packname for the given user.
        This means, that the gold cost of the pack is removed from the users inventory and a pack of that type is added
        If the user does not have enough gold, this method will raise a PermissionError

        CHANGELOG

        Added 12.06.2019

        :raise: PermissionError

        :param username:
        :param packname:
        :return:
        """
        user = self.get_user(username)
        pack_config = self.CONFIG.PACKS[packname]

        # With this adapter we convert the config information about the pack into a dict, which can be directly
        # unpacked into the model constructor, which is needed by the user.by_pack() method
        parameters_adapter = PackParametersAdapter(packname, pack_config)
        pack_parameters = parameters_adapter.parameters()

        user.buy_pack(pack_parameters)

    def user_get_packs(self, username: str) -> List[Pack]:
        """
        Returns a list of Pack objects from the given users inventory

        CHANGELOG

        Added 12.06.2019

        :param username:
        :return:
        """
        user = self.get_user(username)
        return user.packs

    def user_get_packs_by_name(self, username: str) ->Dict[str, List]:
        """
        Returns a dict, whose keys are the names of different pack types and the values are Pack objects of that type
        from the given users inventory

        CHANGELOG

        Added 14.06.2019

        :param username:
        :return:
        """
        user = self.get_user(username)
        return user.get_packs_by_name()

    # USER REWARD MANAGEMENT
    # ----------------------

    def user_use_reward(self, username: str, rewardname: str):
        """
        Given the type name of one of the rewards in the given users inventory, this method will use the reward for the
        user, which means the effect will be evaluated and the the instance will be deleted.
        If the user does not own a reward of the given type, a LookupError will be risen.

        CHANGELOG

        Added 15.06.2019

        :raise: LookupError

        :param username:
        :param rewardname:
        :return:
        """
        user = self.get_user(username)
        user.use_reward(rewardname)

    def user_buy_reward(self, username: str, rewardname: str):
        """
        Given the name of an available reward this method will buy one instance of that type for the given user.
        If the there is no reward type of the given name, the method will raise a KeyError.
        If the user does not have enough dust to buy that reward, this method will raise a PermissionError

        CHANGELOG

        Added 15.06.2019

        :raise: KeyError, PermissionError

        :param username:
        :param rewardname:
        :return:
        """
        user = self.get_user(username)

        reward_config = self.CONFIG.REWARDS[rewardname]
        # The "user.buy_reward" function expects a complete parameter dict for the Reward class, this has to be created
        # first by plugging the dict from the config file for the reward type into the parameters adapter.
        # Every parameters adapter conforms to an interface, where the "parameters" method returns the parameters dict
        parameters_adapter = RewardParametersAdapter(rewardname, reward_config)
        reward_parameters = parameters_adapter.parameters()

        user.buy_reward(reward_parameters)

    def user_recycle_reward(self, username: str, rewardname: str):
        """
        This method will recylce one of the rewards with the given reward name in the given users inventory. This means
        that the instance of the reward will be removed from the inventory and recycled dust will be added to the inv.
        This method will raise a LookupError, when the user does not own a reward of the given type

        CHANGELOG

        Added 15.06.2019

        :raise: LookupError

        :param username:
        :param rewardname:
        :return:
        """
        user = self.get_user(username)
        user.recycle_reward(rewardname)

    def user_get_rewards(self, username: str) -> List[Reward]:
        """
        Returns a list of Reward objects from the given users inventory

        CHANGELOG

        Added 12.06.2019

        :param username:
        :return:
        """
        user = self.get_user(username)
        return user.rewards

    def user_get_rewards_by_name(self, username: str) -> Dict[str, List]:
        """
        Returns a dict, whose keys are the names of the reward types and the values the Reward objects from the users
        inventory of that type

        CHANGELOG

        Added 14.06.2019

        :param username:
        :return:
        """
        user = self.get_user(username)
        return user.get_rewards_by_name()

    # USER CURRENCY MANAGEMENT
    # ------------------------

    def user_get_gold(self, username: str) -> int:
        """
        Returns the integer amount of gold in the given users inventory

        CHANGELOG

        Added 12.06.2019

        :param username:
        :return:
        """
        user = self.get_user(username)
        return user.gold

    def user_get_dust(self, username: str) -> int:
        """
        Returns the integer amount of dust in the given users inventory

        CHANGELOG

        Added 12.06.2019

        :param username:
        :return:
        """
        user = self.get_user(username)
        return user.dust

    def user_add_gold(self, username: str, amount: int):
        """
        Adds the given amount of gold to the given users inventory

        CHANGELOG

        Added 12.06.2019

        :param username:
        :param amount:
        :return:
        """
        user = self.get_user(username)
        user.gold += amount
        user.save()

    def user_add_dust(self, username: str, amount: int):
        """
        Adds the given amount of dust to the given users inventory

        CHANGELOG

        Added 12.06.2019

        :param username:
        :param amount:
        :return:
        """
        user = self.get_user(username)
        user.dust += amount
        user.save()

    # BACKEND OPERATIONS
    # ------------------

    def backend_update(self):
        """
        This method performs a backend query to update the earned gold of the users.
        It does so by instantiating a new backend object (using the class specified in the config BACKEND) and calling
        the get_update() method on it

        CHANGELOG

        Added 13.06.2019

        :return:
        """
        # The backend class to be used is set in the BACKEND attribute of the config.
        # Every backend class must implement the interface "get_update()", which will return a dict, whose keys are
        # the user names and the values are lists with dicts, each describing an action, that has earned them gold
        backend: AbstractBackend = self.CONFIG.BACKEND()
        action_dict = backend.get_update()
        for username, actions in action_dict.items():
            user = self.get_user(username)
            for action in actions:
                user.gold += action['gold']
            user.save()

    # CONFIGURATION MANAGEMENT
    # ------------------------

    def load(self):
        """
        This method loads all the config from the "config.py" file into the EnvironmentConfig object and initializes
        the database.
        NOTE: This method has to be called as the first thing in every programs execution! All following steps
        require this.

        CHANGELOG

        Added 10.06.2019

        :return:
        """
        # This method will issue the EnvironmentConfig object to read the config.py file to save all the variables
        # defined in there as class variables. This means all variables defined in config.py can be accessed through
        # the config object
        self.CONFIG.load()

        # The "init" method creates a new database object, based on what engine is specified.
        # The "connect" method will issue the database object to connect to the actual persistent representation of the
        # database (file for sqlite and server for mysql)
        self.CONFIG.database_config.init()
        self.CONFIG.database_config.connect()

    # SETTING UP / INSTALLING
    # -----------------------

    def create_tables(self):
        """
        Will create the necessary tables within the database to correctly represent the rewardify models.
        NOTE: This should only be called when the database was first created

        CHANGELOG

        Added 10.06.2019

        :return:
        """
        # All the models defined for this project are connected to the database object and this method will make the
        # database create all tables, needed to correctly represent the models
        self.CONFIG.database_config.create_tables()

    # HELPER METHODS
    # --------------

    def available_rewards_parameters(self):
        """
        This method will return a list of the parameter dicts of all the available rewards

        CHANGELOG

        Added 12.06.2019

        :return:
        """
        rewards = []
        for name, config in self.CONFIG.REWARDS.items():
            parameter_adapter = RewardParametersAdapter(name, config)
            rewards.append(parameter_adapter.parameters())

        return rewards



# standard library
import datetime

from typing import Dict

# local imports
from rewardify.rarity import Rarity


# #######################
# INTERFACE SPECIFICATION
# #######################


class ModelParameterAdapterInterface:
    """
    This is an interface specifying, what a model parameter adapter has to implement.
    First of all: What is the use case of a "model parameter adapter"? Take the example of possible rewards. To the
    users available rewards are defined as a dict like structure in the config file, the keys of these dicts do not
    correlate one to one to the names of the model fields though and some model fields might even need to be computed
    from other values. This is where an adapter comes in. The dict from the config is plugged in and through
    the "parameters()" method of the adapter the complete parameters dict for the model is created, with which a model
    can be instantiated like this: "model = Model( **adapter.parameters() )"

    CHANGELOG

    Added 12.06.2019
    """

    def parameters(self) -> Dict:
        """
        This is the only method, that has to be implemented. It has to return a dict, where the keys are exactly named
        like the fields of the models and the values being the values, that were passed to the constructor of the
        adapter.

        CHANGELOG

        Added 12.06.2019

        :return:
        """
        raise NotImplementedError


# ################
# BEHAVIOUR MIXINS
# ################


class SlugAdapterMixin:
    """
    This is a behaviour mixin for parameter adapter implementations. It contains the common functionality for creating
    the slug from the plain text name of a model object.

    CHANGELOG

    Added 12.06.2019
    """
    @classmethod
    def create_slug(cls, name) -> str:
        """
        Given the name of an instance, this method will create the slug from that name, by putting it into all lower
        case and replacing the whitespaces with underscores.

        CHANGELOG

        Added 12.06.2019

        :param name:
        :return:
        """
        # The slug is being created to have a more processing friendly version of the name, which means all lower case
        # and no white spaces (these are being replaced by underscores)
        slug = name.lower()
        slug = slug.replace(' ', '_')

        return slug


# ###################
# THE ACTUAL ADAPTERS
# ###################


class RewardParametersAdapter(SlugAdapterMixin, ModelParameterAdapterInterface):
    """
    This adapter object takes the name of an available reward and its configuration dict as specified in the config
    file of the project and created a parameters dict upon calling the "parameters()" method which can be directly
    unpacked into the constructor of the Reward model class

    CHANGELOG

    Added 12.06.2019

    Changed 15.06.2019
    Made adjustments for the "effect" parameter of the Reward class in the new version
    """

    # The keys are the string names, which the user can use in the config file and the values are the constants, which
    # the code uses internally to create a new Reward object.
    RARITY_MAPPING = {
        'common':           Rarity.COMMON,
        'uncommon':         Rarity.UNCOMMON,
        'rare':             Rarity.RARE,
        'legendary':        Rarity.LEGENDARY,
    }

    # 15.06.2019
    # This default dict is needed as the effect parameter for example is OPTIONAL, which means, it does not appear in
    # every config dict for a reward, but we have to provide a default value to the database model anyways
    DEFAULT_CONFIG = {
        'description':      '',
        'rarity':           'common',
        'effect':           '',
        'cost':             0,
        'recycle':          0
    }

    def __init__(self, name: str, config: Dict):
        """
        The constructor.

        CHANGELOG

        Added 12.06.2019

        Changed 15.06.2019
        The config dict for the reward is now first being initialized with the DEFAULT_CONFIG dict and then updated
        with the actual values of the given config.
        This is to provide a default for eventually not given optional parameters (such as "effect")

        :param name:
        :param config:
        """
        self.name = name
        self.config = self.DEFAULT_CONFIG
        self.config.update(config)

    def parameters(self) -> Dict:
        """
        Returns the parameters dict for the constructor of the Reward model class

        CHANGELOG

        Added 12.06.2019

        Changed 15.06.2019
        Added the "effect" parameter, which has been introduced with the new version. The effect parameter will contain
        a string directive, which denotes which effect the reward has on the system. For example granting dust

        :return:
        """
        parameters = {
            'name':             self.name,
            'slug':             self.create_slug(self.name),
            'description':      self.config['description'],
            'rarity':           self.convert_rarity(self.config['rarity']),
            'effect':           self.config['effect'],
            'dust_cost':        self.config['cost'],
            'dust_recycle':     self.config['recycle'],
            'date_obtained':    datetime.datetime.now()
        }
        return parameters

    def convert_rarity(self, rarity: str):
        """
        Given the string name of the rarity from the config dict, this method will return the according constant,
        that will be recognized by the field as a valid identifier for one of the four rarities

        CHANGELOG

        Added 12.06.2019

        :param rarity:
        :return:
        """
        return self.RARITY_MAPPING[rarity]


class PackParametersAdapter(SlugAdapterMixin, ModelParameterAdapterInterface):
    """
    This adapter object takes the name of an available pack and its configuration dict as specified in the config
    file of the project and created a parameters dict upon calling the "parameters()" method which can be directly
    unpacked into the constructor of the Pack model class

    CHANGELOG

    Added 12.06.2019
    """

    def __init__(self, name: str, config: Dict):
        self.name = name
        self.config = config

    def parameters(self):
        """
        Returns the parameters dict for the constructor of the Pack model class

        CHANGELOG

        Added 12.06.2019

        :return:
        """
        parameters = {
            'name':             self.name,
            'slug':             self.create_slug(self.name),
            'description':      self.config['description'],
            'gold_cost':        self.config['cost'],
            'slot1':            self.config['1'],
            'slot2':            self.config['2'],
            'slot3':            self.config['3'],
            'slot4':            self.config['4'],
            'slot5':            self.config['5'],
            'date_obtained':    datetime.datetime.now()
        }
        return parameters



# standard library
import datetime

from unittest import TestCase

# local imports
from rewardify._util_test import DBTestCase

from rewardify.adapters import PackParametersAdapter, RewardParametersAdapter

from rewardify.rarity import Rarity


class TestAdaptersRaw(TestCase):

    def test_pack_adapter(self):
        # These are the informations, that would be given to the adapter coming from the config file
        name = 'Misc Pack'
        config = {
            'description':          'this is only for testing purposes',
            'cost':                 100,
            '1':                    [1, 0, 0, 0],
            '2':                    [1, 0, 0, 0],
            '3':                    [1, 0, 0, 0],
            '4':                    [1, 0, 0, 0],
            '5':                    [1, 0, 0, 0]
        }

        adapter = PackParametersAdapter(name, config)
        parameters = adapter.parameters()

        self.assertEqual(parameters['name'], name)
        self.assertListEqual(parameters['slot1'], [1, 0, 0, 0])
        self.assertIsInstance(parameters['date_obtained'], datetime.datetime)
        self.assertEqual(parameters['gold_cost'], 100)

    def test_reward_adapter(self):
        # These are the informations, that would be given to the adapter coming from the config file
        name = 'Misc Reward'
        config = {
            'description':          'this is only for testing purposes',
            'rarity':               'common',
            'recycle':              500,
            'cost':                 1000,
        }

        adapter = RewardParametersAdapter(name, config)
        parameters = adapter.parameters()

        self.assertEqual(parameters['name'], name)
        self.assertEqual(parameters['rarity'], Rarity.COMMON)
        self.assertEqual(parameters['dust_cost'], 1000)
        self.assertIsInstance(parameters['date_obtained'], datetime.datetime)



# Standard library
from typing import Union, List

# Third party
import numpy as np

from peewee import Field

# Local imports
from rewardify.rarity import Rarity


class RarityProbability:
    """
    Instances of this class represent a probability distribution for the different rarity values. These probab. values
    denote the probability, that a reward of this rarity will appear from a pack.

    CHANGELOG

    Added 09.06.2019
    """
    def __init__(self,
                 common_probability: float,
                 uncommon_probability: float,
                 rare_probability: float,
                 legendary_probability: float):
        self.dict = {
            Rarity.COMMON:      common_probability,
            Rarity.UNCOMMON:    uncommon_probability,
            Rarity.RARE:        rare_probability,
            Rarity.LEGENDARY:   legendary_probability
        }

    # METHODS
    # -------

    def choice(self):
        """
        This method will make an actual random decision and return a rarity identifier (from Rarity.RARITIES) based on
        the given probability distribution of this object.

        CHANGELOG

        Added 12.06.2019

        :return:
        """
        result = np.random.choice(list(self.dict.keys()), 1, p=list(self.dict.values()))
        return result[0]

    def list(self) -> List[float]:
        """
        This method returns a list, which contains the probability values for the different rarities in ascending order
        of rarity magnitude.

        CHANGELOG

        Added 09.06.2019

        :return:
        """
        return list(self.dict.values())

    def csv(self, accuracy: int = 4) -> str:
        """
        This methods returns a string, where the four probability values are separated by commas, each value having as
        many places after the decimal point as given by the "accuracy" argument.

        EXAMPLE:
        accuracy=4
        >> "0.5000,0.2000,0.2000,0.1000"

        CHANGELOG

        Added 09.06.2019

        :param accuracy:
        :return:
        """
        # First we create a template string for converting the float numbers for each probability into strings with the
        # amount of numbers after the decimal point as specified by "accuracy"
        template = "{%s}" % ":01.{}f".format(accuracy)
        format_string = ','.join([template, template, template, template])

        # Now we use the the probability values to fill in the format string
        probabilities = self.list()
        csv_string = format_string.format(*probabilities)
        return csv_string

    def equals(self, prob):
        """
        Returns true, if this object describes the same probability distribution as the other RarityProbability object
        "prob" given (which means all the values have to be the same). Returns False otherwise

        CHANGELOG

        Added 09.06.2019

        :param prob:
        :return:
        """
        for key, value in self.dict.items():
            # If any value is found, that is not the same, the two objects are not the same
            if prob[key] != value:
                return False

        # If the above loop exited, without returning False prematurely, the two objects are equal
        return True

    # PROPERTIES
    # ----------

    # MAGIC METHODS
    # -------------

    def __len__(self):
        return len(self.dict.keys())

    def __getitem__(self, item):
        """
        This method is being called, when an "Rarity Probability" is being indexed with "item".
        It can be indexed by all the means of specifiying a rarity: correct string or int id or a Rarity object
        directly

        CHANGELOG

        Added 09.06.2019

        :param item:
        :return:
        """
        rarity = Rarity(item)
        return self.dict[str(rarity)]

    def __setitem__(self, key, value):
        """
        It can be indexed by all the means of specifiying a rarity: correct string or int id or a Rarity object
        directly

        CHANGELOG

        Added 09.06.2019

        :param key:
        :param value:
        :return:
        """
        rarity = Rarity(key)
        self.dict[str(rarity)] = float(value)

    def __eq__(self, other):
        """
        Returns true, if this object describes the same probability distribution as the other RarityProbability object
        "prob" given (which means all the values have to be the same). Returns False otherwise.
        Raises a TypeError in case one should try to compare it with any other type of object.

        CHANGELOG

        Added 09.06.2019

        :raises: TypeError

        :param other:
        :return:
        """
        if isinstance(other, RarityProbability):
            return self.equals(other)
        else:
            raise TypeError('You cannot compare a RarityProbability with anything else than objects of same type!')

    # CLASS METHODS
    # -------------

    @classmethod
    def from_iterable(cls, probabilities):
        """
        Creates a new RarityProbability object from an iterable (list or tuple), where the order of the elements is
        being interpreted as the order of the rarities in ascending order

        CHANGELOG

        Added 09.06.2019

        :param probabilities:
        :return:
        """
        return RarityProbability(*probabilities)

    @classmethod
    def from_csv(cls, string: str):
        """
        Creates new RarityProbability object from a csv string.

        CHANGELOG

        Added 09.06.2019

        :param string:
        :return:
        """
        # First we obviously get a lost of the values by splitting by the commas
        values = string.split(',')

        # These values of the list are still strings though and still need to be converted into the
        # floats
        float_values = list(map(float, values))

        # Creating a new object using these values
        return RarityProbability(*float_values)


class ProbabilityField(Field):
    """
    This is a custom field for the database models, which represents the probability distribution for the appearance of
    the different rarities of rewards in a pack slot.

    CHANGELOG

    Added 09.06.2019
    """
    # CLASS VARIABLES
    # ---------------

    # This is the error message, that is being shown, when the value of a ProbabilityField is set with a value
    # of an unsupported type
    DB_VALUE_ERROR = 'You need to pass either a list or a RarityProbability object to the ProbabilityField!'

    # This is the type of column to which this field maps in an actual database
    field_type = 'text'

    def python_value(self, value):
        # For getting the python value, we just have to create a RarityProbability from the csv text saved in the
        # database
        return RarityProbability.from_csv(value)

    def db_value(self, value):
        # First of all the value needs to be either a list, that contains 4 floats to specify the probabilities or a
        # RarityProbability object directly.
        # If it is a list we create a new RarityProbability object from that list first
        rarity_probability = self.convert_rarity_probability(value)

        # Now we create a csv string from the object, which can be saved into the text column
        return rarity_probability.csv()

    # HELPER METHODS
    # --------------

    def convert_rarity_probability(self,
                                   value: Union[RarityProbability, List[float]]
                                   ) -> RarityProbability:
        """

        :param value:
        :return:
        """
        if isinstance(value, list):
            probability = RarityProbability.from_iterable(value)
        elif isinstance(value, RarityProbability):
            probability = value
        else:
            raise ValueError(self.DB_VALUE_ERROR)

        return probability



# Standard library
import datetime

from typing import List, Dict, Any

# local imports
from rewardify.env import EnvironmentConfig

from rewardify.backends.base import AbstractBackend, CheckConfigMixin


class MockBackend(CheckConfigMixin, AbstractBackend):
    """
    This is a mock object for testing purposes. It implements the AbstractBackend.
    For every call to the get_update() method it will return a list of exactly one action granting 100
    gold to every user, that is specified in the config variable MOCK_BACKEND_USERS.

    CHANGELOG

    Added 10.09.2019

    Changed 13.06.2019
    Removed the method "check_config" and created the "CheckConfigMixin" from that. This class now inherits from this
    mixin, thus retaining the method.
    """

    # THE CONFIG ACCESS
    # -----------------

    # This is a list of all the variable names, this backend expects to be defined in the config file to the program
    EXPECTED_CONFIG = [
        'MOCK_BACKEND_USERS'
    ]

    # A reference to the config singleton
    CONFIG: EnvironmentConfig = EnvironmentConfig.instance()

    # CLASS CONSTANTS
    # ---------------

    # This is the action description dict, which will be given to every user on the call of the
    # "get_update" method
    MOCK_ACTION = {
        'name':             'Mockery',
        'description':      'For doing absolutely nothing!',
        'gold':             100,
        'date':             datetime.datetime.now()
    }

    # INSTANCE CONSTRUCTION
    # ---------------------

    def __init__(self):
        """
        The constructor.

        CHANGELOG

        Added 10.09.2019
        """
        # This method will check if all the variables expected from the config file are actually defined in the config
        # file and will raise an exception, if that is not the case.
        self.check_config()

    def get_update(self):
        """
        This method returns the update dict as defined by the abstract base class

        CHANGELOG

        Added 10.06.2019

        :return:
        """
        update_dict = {}
        for user in self.CONFIG.MOCK_BACKEND_USERS:
            update_dict[user] = [self.MOCK_ACTION]
        return update_dict

    # IMPLEMENTING MIXIN REQUIREMENTS
    # -------------------------------

    def get_config(self):
        """
        Returns an instance of the EnvironmentConfig instance

        CHANGELOG

        Added 13.06.2019

        :return:
        """
        return self.CONFIG

    def expected_config(self):
        """
        Returns a list with the string names of the expected variables from the config object

        CHANGELOG

        Added 13.06.2019

        :return:
        """
        return self.EXPECTED_CONFIG



# Standard library
from unittest import TestCase

# Third party imports
from peewee import Model
from peewee import SqliteDatabase
from peewee import TextField

# Local imports
from rewardify.password import PasswordHash
from rewardify.password import PasswordField
from rewardify.password import VerifyMismatchError


class PasswordTest(TestCase):

    def test_hashing_works(self):
        password_string = "super secret"

        password_hash = PasswordHash(password_string)
        self.assertTrue(password_hash.check(password_string))

    def test_password_verification_works(self):
        password_string = "super secret"
        wrong_password_string = "lucky guess?"

        password_hash = PasswordHash(password_string)
        self.assertRaises(
            VerifyMismatchError,
            password_hash.verify,
            wrong_password_string
        )

    def test_password_hash_recognition_works(self):
        password_string = "super secret"
        password_hash = PasswordHash(password_string)

        self.assertFalse(PasswordHash.is_hash(password_string))
        self.assertTrue(PasswordHash.is_hash(password_hash))

    def test_instantiating_password_hash_with_hashed_string_works(self):
        # First we create the actual hash
        password_string = "super secret"
        password_hash = PasswordHash(password_string)

        # Then we create a new PasswordHash object from the already hashed string
        hashed_password_string = str(password_hash)
        hashed_password_hash = PasswordHash(hashed_password_string)

        # In theory the second constructor should not modify the already hashed string, thus they be the same.
        self.assertEqual(str(password_hash), str(hashed_password_hash))


# TESTING THE ACTUAL PASSWORD FIELD
# ---------------------------------


class PasswordFieldModel(Model):
    """
    This is a dummy model simply to test the working of the password field
    """
    name = TextField()
    password = PasswordField()


class PasswordFieldTest(TestCase):

    TEST_DATABASE = SqliteDatabase(':memory:')
    MODELS = [PasswordFieldModel]

    # SETTING UP THE TEST DATABASE
    # ----------------------------

    def setUp(self):
        self.TEST_DATABASE.bind(self.MODELS, bind_refs=False, bind_backrefs=False)

        self.TEST_DATABASE.connect()
        self.TEST_DATABASE.create_tables(self.MODELS)

    def tearDown(self):
        self.TEST_DATABASE.drop_tables(self.MODELS)
        self.TEST_DATABASE.close()

    # ACTUAL TEST METHODS
    # -------------------

    def test_password_field_works(self):
        name_string = 'test_password_field_works'

        # Here we create the model and save it to the database
        password_string = "super secret"
        model = PasswordFieldModel(name=name_string, password=password_string)
        model.save()

        # Now we load it and check if it has worked
        model = PasswordFieldModel.get(PasswordFieldModel.name == name_string)
        self.assertTrue(model.password.check(password_string))




# standard library
import unittest
import os
import datetime
import json

# third party
import pandas
import pandas.api.types as ptypes

# local imports
from rewardify._util_test import RewardifyTestCase

from rewardify.backends import MockBackend, ForestBackend
from rewardify.backends.forest import ForestCSVFile

from rewardify.env import EnvironmentConfig


# #################################################################
# TESTING THE GENERAL INTERFACE FUNCTIONALITY WITH THE MOCK BACKEND
# #################################################################


class TestMockBackend(RewardifyTestCase):

    CONFIG: EnvironmentConfig = EnvironmentConfig.instance()

    SAMPLE_IMPORTS = [
        'from rewardify.backends import AbstractBackend'
    ]

    SAMPLE_DATABASE_DICT = {
        'engine': 'sqlite',
        'host': ':memory:',
        'database': '',
        'user': '',
        'password': '',
        'port': 0
    }

    SAMPLE_PACKS = [
        {
            'name':         'SamplePack',
            'description':  'shitty pack',
            'cost':         1000,
            'slot1':        [1, 0, 0, 0],
            'slot2':        [1, 0, 0, 0],
            'slot3':        [1, 0, 0, 0],
            'slot4':        [1, 0, 0, 0],
            'slot5':        [1, 0, 0, 0],
        }
    ]

    SAMPLE_REWARDS = [
        {
            'name':         'SampleReward',
            'cost':         1000,
            'recycle':      1000,
            'description':  'none',
            'rarity':       'common'
        }
    ]

    SAMPLE_BACKEND = 'MockBackend'

    SAMPLE_PLUGIN_CODE = 'MOCK_BACKEND_USERS = ["Jonas"]'

    def test_mock_creation_works(self):
        self.setup_sample()

        mock_backend = MockBackend()
        update_dict = mock_backend.get_update()
        self.assertTrue('Jonas' in update_dict.keys())
        self.assertDictEqual(MockBackend.MOCK_ACTION, update_dict['Jonas'][0])

    # HELPER METHODS
    # --------------

    def setup_sample(self):
        # Setting up the config-py file
        self.create_config(
            self.SAMPLE_IMPORTS,
            self.SAMPLE_DATABASE_DICT,
            self.SAMPLE_BACKEND,
            self.SAMPLE_REWARDS,
            self.SAMPLE_PACKS,
            self.SAMPLE_PLUGIN_CODE
        )


# ##################
# FOREST APP BACKEND
# ##################

class TestForestCSVFile(RewardifyTestCase):

    CONFIG: EnvironmentConfig = EnvironmentConfig.instance()

    SAMPLE_CSV_NAME = 'sample.csv'

    SAMPLE_CSV_CONTENT = '''"Start Time","End Time","Tag","Note","Tree Type","Is Success"
"Mon Dec 10 09:43:41 GMT+01:00 2018","Mon Dec 10 09:53:41 GMT+01:00 2018","Nicht gesetzt","","Busch","True"
"Mon Dec 10 09:54:23 GMT+01:00 2018","Mon Dec 10 10:24:23 GMT+01:00 2018","Nicht gesetzt","","Zeder","True"
"Mon Dec 10 10:54:11 GMT+01:00 2018","Mon Dec 10 11:24:11 GMT+01:00 2018","Nicht gesetzt","","Zeder","True"'''

    # ACTUAL TESTS
    # ------------

    def test_sample_csv_content_alignment(self):
        # We check if a TAB is in the sample string. Because i am not sure how that works with the indentation of the
        # class variable and a multiline string.
        self.assertTrue("    " not in self.SAMPLE_CSV_CONTENT)

    def test_csv_file_is_being_created(self):
        self.create_csv()
        # Checking if the file exists
        self.assertTrue(os.path.exists(self.csv_path()))
        # Content correct?
        with open(self.csv_path(), mode='r') as file:
            content = file.read()
            self.assertEqual(content, self.SAMPLE_CSV_CONTENT)

    def test_load_file_pandas_data_types(self):
        self.create_csv()

        csv_file = ForestCSVFile(self.csv_path())
        # Making sure the datetime fields are parsed as such
        self.assertTrue(ptypes.is_datetime64_ns_dtype(csv_file.data_frame.dtypes['Start Time']))
        self.assertTrue(ptypes.is_datetime64_ns_dtype(csv_file.data_frame.dtypes['End Time']))
        # Checking if the boolean "is Success" column is actually boolean
        self.assertTrue(ptypes.is_bool_dtype(csv_file.data_frame.dtypes['Is Success']))

    def test_calculating_duration(self):
        self.create_csv()

        csv_file = ForestCSVFile(self.csv_path())
        for index, row in csv_file.data_frame.iterrows():
            duration = csv_file.calculate_duration(row['Start Time'], row['End Time'])
            duration_calculated = int((row['End Time'] - row['Start Time']).seconds / 60)
            self.assertEqual(duration, duration_calculated)

    def test_get_dicts(self):
        self.create_csv()

        csv_file = ForestCSVFile(self.csv_path())

        # The dicts, which are returned by the "get_dicts" method are supposed to have to values, one for the
        # tag of the tree (which is important for how much gold per minute is granted) and the duration in minutes,
        # which the activity lasted.
        activity_dicts = csv_file.get_dicts()
        self.assertEqual(len(activity_dicts), 3)
        dict_keys = list(activity_dicts[0].keys())
        self.assertListEqual(dict_keys, ['tag', 'duration'])

    def test_get_dicts_after(self):
        self.create_csv()
        csv_file = ForestCSVFile(self.csv_path())

        # This is the date time of the second tree in the sample CSV. Which means only the last tree should turn up in
        # the list using the "_after" method!
        date_time = pandas.to_datetime("Mon Dec 10 09:54:23 GMT+01:00 2018")
        activity_dicts = csv_file.get_dicts_after(date_time)
        self.assertEqual(len(activity_dicts), 1)

    # HELPER METHODS
    # --------------

    def create_csv(self):
        path = self.csv_path()
        with open(path, mode='w+') as file:
            file.write(self.SAMPLE_CSV_CONTENT)

    def csv_path(self):
        return os.path.join(self.FOLDER_PATH, self.SAMPLE_CSV_NAME)


class TestForestBackend(RewardifyTestCase):

    # SAMPLE CONFIG FILE SETUP
    # ------------------------

    SAMPLE_IMPORTS = [
        'from rewardify.backends.forest import ForestBackend'
    ]

    SAMPLE_DATABASE_DICT = {
        'engine':   'sqlite',
        'host':     ':memory:',
        'port':     0
    }

    SAMPLE_PACKS = []
    SAMPLE_REWARDS = []

    SAMPLE_BACKEND = 'ForestBackend'

    SAMPLE_PLUGIN_CODE = '''
FOREST_BACKEND_PATH = "/tmp/rewardify"
FOREST_BACKEND_SETTINGS = {
    "Jonas": {
        "name":     "Jonas",
        "period":   60,
        "default":  1000
    },
    "Joana": {
        "name":     "LuaMia",
        "period":   30,
        "default":  10     
    },
}'''

    # SAMPLE FOREST SETUP
    # -------------------

    SAMPLE_CSV_FILES = {
        'Plants of jonseb1998@gmail.com (Jonas).csv': '''"Start Time","End Time","Tag","Note","Tree Type","Is Success"
"Mon Dec 10 16:01:22 GMT+01:00 2018","Mon Dec 10 16:31:22 GMT+01:00 2018","Nicht gesetzt","","Zeder","True"
"Mon Dec 10 20:01:08 GMT+01:00 2018","Mon Dec 10 20:31:08 GMT+01:00 2018","Nicht gesetzt","","Zeder","True"
"Tue Dec 11 08:15:04 GMT+01:00 2018","Tue Dec 11 09:00:04 GMT+01:00 2018","Nicht gesetzt","","Zeder","True"
"Tue Dec 11 09:16:48 GMT+01:00 2018","Tue Dec 11 09:46:48 GMT+01:00 2018","Nicht gesetzt","","Zeder","True"
"Tue Dec 11 10:01:44 GMT+01:00 2018","Tue Dec 11 10:41:44 GMT+01:00 2018","Nicht gesetzt","","Zeder","True"
"Tue Dec 11 11:17:17 GMT+01:00 2018","Tue Dec 11 11:47:17 GMT+01:00 2018","Nicht gesetzt","","Zeder","True"
        ''',
        'Plants of luamia@gmail.com (LuaMia).csv': '''"Start Time","End Time","Tag","Note","Tree Type","Is Success"
"Fri Dec 14 07:37:54 GMT+01:00 2018","Fri Dec 14 08:07:54 GMT+01:00 2018","Nicht gesetzt","","Zeder","True"
"Fri Dec 14 08:29:21 GMT+01:00 2018","Fri Dec 14 08:59:21 GMT+01:00 2018","Nicht gesetzt","","Zeder","True"
"Fri Dec 14 09:01:32 GMT+01:00 2018","Fri Dec 14 09:31:32 GMT+01:00 2018","Nicht gesetzt","","Zeder","True"
"Fri Dec 14 10:13:53 GMT+01:00 2018","Fri Dec 14 10:43:53 GMT+01:00 2018","Nicht gesetzt","","Zeder","True"
"Fri Dec 14 15:12:35 GMT+01:00 2018","Fri Dec 14 15:42:35 GMT+01:00 2018","Nicht gesetzt","","Zeder","True"
"Sat Dec 15 09:08:42 GMT+01:00 2018","Sat Dec 15 09:18:01 GMT+01:00 2018","Nicht gesetzt","","Zeder","False"
"Sat Dec 15 09:19:09 GMT+01:00 2018","Sat Dec 15 10:14:09 GMT+01:00 2018","Nicht gesetzt","","Zeder","True"
        '''
    }

    # ACTUAL TESTS
    # ------------

    def test_forest_backend_test_case_generally_working(self):
        self.setup_sample()

        forest_backend = ForestBackend()
        csv_paths = self.get_sample_csv_paths()
        for path in csv_paths:
            self.assertTrue(os.path.exists(path))

    def test_folder_path_correct(self):
        self.setup_sample()

        forest_backend = ForestBackend()
        self.assertEqual(forest_backend.folder_path, '/tmp/rewardify')

    def test_loading_csv_files(self):
        self.setup_sample()

        forest_backend = ForestBackend()
        self.assertEqual(len(forest_backend.csv_files), 2)
        self.assertIsInstance(forest_backend.csv_files[0], ForestCSVFile)

    def test_extract_forest_name_from_csv_path(self):
        self.setup_sample()

        csv_paths = self.get_sample_csv_paths()
        forest_names = list(map(ForestBackend.forest_name_from_path, csv_paths))
        self.assertListEqual(forest_names, ['Jonas', 'LuaMia'])

    def test_create_forest_user_map(self):
        self.setup_sample()

        forest_backend = ForestBackend()
        # This method should return a dict, whose keys are the string user names for forest and the values are the
        # string user names for rewardify
        forest_user_map = forest_backend.get_forest_user_mapping()

        expected = {
            'Jonas':    'Jonas',
            'LuaMia':   'Joana'
        }
        self.assertDictEqual(forest_user_map, expected)

    def test_default_state_works(self):
        self.setup_sample()

        forest_backend = ForestBackend()
        # This method is supposed to return a dict, whose keys are the string user names of the rewardify users and
        # the values pandas Datetime objects for some time 1701. This is supposed to represent the last tree, the
        # backend has granted gold for.
        state = forest_backend._default_state()

        self.assertIsInstance(state, dict)
        self.assertEqual(len(state), 2)
        self.assertEqual(state['Jonas'].year, 1701)

    def test_load_state_works(self):
        self.setup_sample()

        # Now we also need to create the json file, which contains the state
        date_time = pandas.to_datetime("Mon Dec 10 16:01:22 GMT+01:00 2018")
        state = {
            'Jonas':        str(date_time),
            'Joana':        str(date_time)
        }
        file_path = os.path.join(self.FOLDER_PATH, ForestBackend.STATE_FILE_NAME)
        with open(file_path, mode="w") as file:
            json.dump(state, file)

        self.assertTrue(os.path.exists(file_path))

        forest_backend = ForestBackend()
        self.assertNotEqual(forest_backend.state['Jonas'].year, 1701)

    def test_get_update(self):
        self.setup_sample()

        forest_backend = ForestBackend()
        update_dict = forest_backend.get_update()
        self.assertEqual(len(update_dict), 2)
        self.assertEqual(len(update_dict['Joana']), 7)
        self.assertEqual(len(update_dict['Jonas']), 6)

    # HELPER METHODS
    # --------------

    def setup_sample(self):
        # This method will use a template and the given values to create a new "config.py" file in the testing config
        # folder
        self.create_config(
            self.SAMPLE_IMPORTS,
            self.SAMPLE_DATABASE_DICT,
            self.SAMPLE_BACKEND,
            self.SAMPLE_REWARDS,
            self.SAMPLE_PACKS,
            self.SAMPLE_PLUGIN_CODE
        )

        # This method will create all the CSV files into the testing config folder
        self.create_sample_csv_files()

    def create_sample_csv_files(self):
        for name, content in self.SAMPLE_CSV_FILES.items():
            path = os.path.join(self.FOLDER_PATH, name)
            with open(path, mode='w') as file:
                file.write(content)

    def get_sample_csv_paths(self):
        paths = []
        for name in self.SAMPLE_CSV_FILES.keys():
            path = os.path.join(self.FOLDER_PATH, name)
            paths.append(path)
        return paths
